#!/usr/bin/python
# -*- coding: utf-8 -*-

try:
    from . import core
    from . import tasks
    from .scripts.Base import Base
    from .core.devices.Device import Device as device
    from .core.devices.Device import DeviceLoad as load
    from .tasks.Poll import SimplePoll as poll
    from .tasks.Match import Match as match
    from .core.utils.notes import update_log_level as log_level
    from .infos import __version__ as version
    
    # To be able to use the complete version pandas, flask and bokeh must be installed.
    try:
        import pandas
        import bokeh
        import flask
        _COMPLETE = True
    except ImportError:
        _COMPLETE = False
    
    if _COMPLETE:
        from .scripts.Complete import Complete as connect
        from .scripts.Lite import Lite as lite
    else:
        from .scripts.Lite import Lite as connect

except ImportError:
    pass # Probably installing the app...


#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
#
# Licensed under LGPLv3, see file LICENSE in this source tree.
"""
Utility function to retrieve a functionnal IP and a correct broadcast IP
address.
Goal : not use 255.255.255.255 as a broadcast IP address as it is not
accepted by every devices (>3.8.38.1 bacnet.jar of Tridium Jace for example)

"""
from bacpypes.pdu import Address

from ..io.IOExceptions import NetworkInterfaceException

import socket
import subprocess
import ipaddress
import sys
import re

class HostIP():
    """
    Special class to identify host IP informations
    """

    def __init__(self):
        ip = self._findIPAddr()
        mask = self._findSubnetMask(ip)
        self.interface = ipaddress.IPv4Interface("%s/%s" % (ip, mask))
    
    @property    
    def ip_address(self):
        """
        IP Address/subnet
        """
        return ('%s/%s' % (self.interface.ip.compressed,
                           self.interface.exploded.split('/')[-1]))

    @property
    def address(self):
        """
        IP Address using bacpypes Address format
        """
        return (Address('%s/%s' % (self.interface.ip.compressed,
                                   self.interface.exploded.split('/')[-1])))

    def _findIPAddr(self):
        """
        Retrieve the IP address connected to internet... used as
        a default IP address when defining Script

        :returns: IP Adress as String
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            s.connect(('google.com', 0))
            addr = s.getsockname()[0]
            print('Using ip : {addr}'.format(addr=addr))
            s.close()
        except socket.error:
            raise NetworkInterfaceException(
                'Impossible to retrieve IP, please provide one manually')
        return addr

    def _findSubnetMask(self, ip):
        """
        Retrieve the broadcast IP address connected to internet... used as
        a default IP address when defining Script

        :param ip: (str) optionnal IP address. If not provided, default to getIPAddr()
        :param mask: (str) optionnal subnet mask. If not provided, will try to find one using ipconfig (Windows) or ifconfig (Linux or MAC)

        :returns: broadcast IP Adress as String
        """
        ip = ip

        if 'win32' in sys.platform:
            try:
                proc = subprocess.Popen('ipconfig', stdout=subprocess.PIPE)
                while True:
                    line = proc.stdout.readline()
                    if ip.encode() in line:
                        break
                mask = proc.stdout.readline().rstrip().split(
                    b':')[-1].replace(b' ', b'').decode()
            except:
                raise NetworkInterfaceException('Cannot read IP parameters from OS')
        else:
            """
            This procedure could use more direct way of obtaining the broadcast IP
            as it is really simple in Unix
            ifconfig gives Bcast directly for example
            or use something like :
            iface = "eth0"
            socket.inet_ntoa(fcntl.ioctl(socket.socket(socket.AF_INET, socket.SOCK_DGRAM), 35099, struct.pack('256s', iface))[20:24])
            """
            pattern = re.compile(r"(255.\d{1,3}.\d{1,3}.\d{1,3})")

            try:
                proc = subprocess.Popen('ifconfig', stdout=subprocess.PIPE)
                while True:
                    line = proc.stdout.readline()
                    if ip.encode() in line:
                        break
                mask = re.findall(pattern,line.decode())[0]
            except:
                mask = '255.255.255.255'
        #self._log.debug('Mask found : %s' %  mask)
        return mask




#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
read_mixin.py - Add ReadProperty and ReadPropertyMultiple to a device 
'''
#--- standard Python modules ---

#--- 3rd party modules ---

#--- this application's modules ---
from ....tasks.Poll import DevicePoll
from ...io.IOExceptions import ReadPropertyMultipleException, NoResponseFromController, SegmentationNotSupported
from ..Points import NumericPoint, BooleanPoint, EnumPoint, OfflinePoint

#------------------------------------------------------------------------------

def retrieve_type(obj_list, point_type_key):
    for point_type, point_address in obj_list:
        if point_type_key in str(point_type):
            yield (point_type, point_address)


class ReadPropertyMultiple():
    """
    Handle ReadPropertyMultiple for a device
    """

    def _batches(self, request, points_per_request):
        """
        Generator for creating 'request batches'.  Each batch contains a maximum of "points_per_request" 
        points to read.
        :params: request a list of point_name as a list
        :params: (int) points_per_request
        :returns: (iter) list of point_name of size <= points_per_request
        """
        for i in range(0, len(request), points_per_request):
            yield request[i:i + points_per_request]


    def _rpm_request_by_name(self, point_list):
        """
        :param point_list: a list of point
        :returns: (tuple) read request for each points, points
        """
        points = []
        requests = []
        for each in point_list:
            str_list = []
            point = self._findPoint(each, force_read=False)
            points.append(point)

            str_list.append(' ' + point.properties.type)
            str_list.append(' ' + str(point.properties.address))
            str_list.append(' presentValue')
            rpm_param = (''.join(str_list))
            requests.append(rpm_param)

        return (requests, points)


    def read_multiple(self, points_list, *, points_per_request=25, discover_request=(None, 6), force_single=False):
        """
        Read points from a device using a ReadPropertyMultiple request.
        [ReadProperty requests are very slow in comparison].

        :param points_list: (list) a list of all point_name as str
        :param points_per_request: (int) number of points in the request

        Requesting many points results big requests that need segmentation.  Aim to request 
        just the 'right amount' so segmentation can be avoided.  Determining the 'right amount' 
        is often trial-&-error.

        :Example:

        device.read_multiple(['point1', 'point2', 'point3'], points_per_request = 10)
        """
        if not self.properties.pss['readPropertyMultiple'] or force_single:
            self._log.warning('Read property Multiple Not supported')
            self.read_single(points_list,points_per_request=1, discover_request=discover_request)
        else:
            if not self.properties.segmentation_supported:
                points_per_request = 1

            if discover_request[0]:
                values = []
                info_length = discover_request[1]
                big_request = discover_request[0]

                for request in self._batches(big_request, points_per_request):
                    
                    try:
                        request = ('{} {}'.format(self.properties.address, ''.join(request)))
                        self._log.debug('RPM_Request: %s ' % request)
                        val = self.properties.network.readMultiple(request)

                        #print('val : ', val, len(val), type(val))
                        if val == None:
                            self.properties.segmentation_supported = False
                            raise SegmentationNotSupported

                    except KeyError as error:
                        raise Exception('Unknown point name : %s' % error)
                        
                    except SegmentationNotSupported as error:
                        self.properties.segmentation_supported = False
                        #self.read_multiple(points_list,points_per_request=1, discover_request=discover_request)
                        self._log.warning('Segmentation not supported')
                
                        self._log.warning('Request too big...will reduce it')
                        if points_per_request == 1:
                            raise
                        self.read_multiple(points_list,points_per_request=1, discover_request=discover_request)
                        
                    else:
                        for points_info in self._batches(val, info_length):
                            values.append(points_info)
                return values

            else:
                big_request = self._rpm_request_by_name(points_list)
                i = 0
                for request in self._batches(big_request[0], points_per_request):
                    try:
                        request = ('{} {}'.format(self.properties.address, ''.join(request)))
                        val = self.properties.network.readMultiple(request)

                    except SegmentationNotSupported as error:
                        self.properties.segmentation_supported = False
                        self.read_multiple(points_list,points_per_request=1, discover_request=discover_request)

                    except KeyError as error:
                        raise Exception('Unknown point name : %s' % error)

                    else:
                        points_values = zip(big_request[1][i:i + len(val)], val)
                        i += len(val)
                        for each in points_values:
                            each[0]._trend(each[1])


    def read_single(self, points_list, *, points_per_request=1, discover_request=(None, 4)):
        if discover_request[0]:
            values = []
            info_length = discover_request[1]
            big_request = discover_request[0]

            for request in self._batches(big_request, points_per_request):
                try:
                    request = ('{} {}'.format(self.properties.address, ''.join(request)))
                    val = self.properties.network.read(request)
    
                except KeyError as error:
                    raise Exception('Unknown point name : %s' % error)
                
                # Save each value to history of each point
                for points_info in self._batches(val, info_length):
                    values.append(points_info)
                    
            return values
        
        else:
            big_request = self._rpm_request_by_name(points_list)
            i = 0
            for request in self._batches(big_request[0], points_per_request):
                try:
                    request = ('{} {}'.format(self.properties.address, ''.join(request)))
                    val = self.properties.network.read(request)
                    points_values = zip(big_request[1][i:i + len(val)], val)
                    
                    i += len(val)
                    for each in points_values:
                        each[0]._trend(each[1])

                except KeyError as error:
                    raise Exception('Unknown point name : %s' % error)


    def _discoverPoints(self, custom_object_list = None):
        if custom_object_list:
            objList = custom_object_list
        else:
            try : 
                    objList = self.properties.network.read(
                    '{} device {} objectList'.format(self.properties.address, self.properties.device_id))
            
            except NoResponseFromController:
                self.log.error('No object list available. Please provide a custom list using the object_list parameter')
                objList = []

            except SegmentationNotSupported:
                objList = []
                number_of_objects = self.properties.network.read(
                    '{} device {} objectList'.format(self.properties.address, self.properties.device_id), arr_index = 0)
    
                for i in range(1,number_of_objects+1):
                    objList.append(self.properties.network.read(
                        '{} device {} objectList'.format(self.properties.address, self.properties.device_id), arr_index = i))

        points = []


        # Numeric
        analog_request = []
        list_of_analog = retrieve_type(objList, 'analog')
        for analog_points, address in list_of_analog:
            analog_request.append('{} {} objectName presentValue units description '.format(analog_points, address))

        try:
            analog_points_info = self.read_multiple('', discover_request=(analog_request, 4), points_per_request=5)
            self._log.info(analog_points_info)
        except SegmentationNotSupported:
            raise

        i = 0
        for each in retrieve_type(objList, 'analog'):
            point_type = str(each[0])
            point_address = str(each[1])
            point_infos = analog_points_info[i]

            if len(point_infos) == 4:
                point_units_state = point_infos[2]
                point_description = point_infos[3]

            elif len(point_infos) == 3:
                #we probably get only objectName, presentValue and units
                point_units_state = point_infos[2]
                point_description = ""

            elif len(point_infos) == 2:
                point_units_state = ""          
                point_description = ""

            else:
                #raise ValueError('Not enough values returned', each, point_infos)
                # SHOULD SWITCH TO SEGMENTATION_SUPPORTED = FALSE HERE
                self._log.warning('Cannot add {} / {} | {}'.format(point_type, point_address, len(point_infos)))
                continue
            
            i += 1
            points.append(
                NumericPoint(
                    pointType=point_type,   pointAddress=point_address, pointName=point_infos[0],
                    description=point_description,  presentValue=float(point_infos[1]), units_state=point_units_state,
                    device=self))

        multistate_request = []
        list_of_multistate = retrieve_type(objList, 'multi')
        for multistate_points, address in list_of_multistate:
            multistate_request.append('{} {} objectName presentValue stateText description '.format(multistate_points, address))

        try:
            multistate_points_info= self.read_multiple('', discover_request=(multistate_request, 4), points_per_request=5)
        except SegmentationNotSupported:
            raise
            
        i = 0
        for each in retrieve_type(objList, 'multi'):
            point_type = str(each[0])
            point_address = str(each[1])
            point_infos = multistate_points_info[i]
            i += 1

            points.append(
                EnumPoint(
                    pointType=point_type,       pointAddress=point_address,     pointName=point_infos[0],
                    description=point_infos[3], presentValue=point_infos[1],    units_state=point_infos[2],
                    device=self))

        binary_request = []
        list_of_binary = retrieve_type(objList, 'binary')
        for binary_points, address in list_of_binary:
            binary_request.append('{} {} objectName presentValue inactiveText activeText description '.format(binary_points, address))

        try:
            binary_points_info= self.read_multiple('', discover_request=(binary_request, 5), points_per_request=5)
        except SegmentationNotSupported:
            raise
            
        i = 0
        for each in retrieve_type(objList, 'binary'):
            point_type = str(each[0])
            point_address = str(each[1])
            point_infos = binary_points_info[i]
            
            if len(point_infos) == 3:
                #we probably get only objectName, presentValue and description
                point_units_state = ('OFF', 'ON')
                point_description = point_infos[2]

            elif len(point_infos) == 5:
                point_units_state = (point_infos[2], point_infos[3])
                point_description = point_infos[4]

                if point_description is None:
                    point_description = ""

            elif len(point_infos) == 2:
                point_units_state = ('OFF', 'ON')          
                point_description = ""

            else:
                #raise ValueError('Not enough values returned', each, point_infos)
                # SHOULD SWITCH TO SEGMENTATION_SUPPORTED = FALSE HERE
                self._log.warning('Cannot add {} / {}'.format(point_type, point_address))
                continue

            i += 1
            points.append(
                BooleanPoint(
                    pointType=point_type,           pointAddress=point_address,     pointName=point_infos[0],
                    description=point_description,  presentValue=point_infos[1],    units_state=point_units_state,
                    device=self))
            
        self._log.info('Ready!')
        return (objList, points)

            
    def poll(self, command='start', *, delay=10):
        """
        Poll a point every x seconds (delay=x sec)
        Can be stopped by using point.poll('stop') or .poll(0) or .poll(False)
        or by setting a delay = 0

        :param command: (str) start or stop polling
        :param delay: (int) time delay between polls in seconds
        :type command: str
        :type delay: int

        :Example:

        device.poll()
        device.poll('stop')
        device.poll(delay = 5)
        """
        if str(command).lower() == 'stop' \
                or command == False \
                or command == 0 \
                or delay == 0:

            if isinstance(self._polling_task.task, DevicePoll):
                self._polling_task.task.stop()
                while self._polling_task.task.is_alive():
                    pass

                self._polling_task.task = None
                self._polling_task.running = False
                self._log.info('Polling stopped')
                
        elif self._polling_task.task is None:
            self._polling_task.task = DevicePoll(self, delay=delay)
            self._polling_task.task.start()
            self._polling_task.running = True
            self._log.info('Polling started, values read every {} seconds'.format(delay))
            
        elif self._polling_task.running:
            self._polling_task.task.stop()
            while self._polling_task.task.is_alive():
                pass
            
            self._polling_task.running = False
            self._polling_task.task = DevicePoll(self, delay=delay)
            self._polling_task.task.start()
            self._polling_task.running = True
            self._log.info('Polling started, every values read each %s seconds' % delay)
            
        else:
            raise RuntimeError('Stop polling before redefining it')


class ReadProperty():
    """
    Handle ReadProperty for a device
    """
    
    def _batches(self, request, points_per_request):
        """
        Generator for creating 'request batches'.  Each batch contains a maximum of "points_per_request" 
        points to read.
        :params: request a list of point_name as a list
        :params: (int) points_per_request
        :returns: (iter) list of point_name of size <= points_per_request
        """
        for i in range(0, len(request), points_per_request):
            yield request[i:i + points_per_request]


    def _rpm_request_by_name(self, point_list):
        """
        :param point_list: a list of point
        :returns: (tuple) read request for each points, points
        """
        points = []
        requests = []
        for each in point_list:
            str_list = []
            point = self._findPoint(each, force_read=False)
            points.append(point)

            str_list.append(' ' + point.properties.type)
            str_list.append(' ' + str(point.properties.address))
            str_list.append(' presentValue')
            rpm_param = (''.join(str_list))
            requests.append(rpm_param)

        return (requests, points)


    def read_multiple(self, points_list, *, points_per_request=1, discover_request=(None, 6)):
        """
        Functions to read points from a device using the ReadPropertyMultiple request.
        Using readProperty request can be very slow to read a lot of data.

        :param points_list: (list) a list of all point_name as str
        :param points_per_request: (int) number of points in the request

        Using too many points will create big requests needing segmentation.
        It's better to use just enough request so the message will not require
        segmentation.

        :Example:

        device.read_multiple(['point1', 'point2', 'point3'], points_per_request = 10)
        """
        #print('PSS : %s' % self.properties.pss['readPropertyMultiple'])
        if isinstance(points_list, list):
            for each in points_list:
                self.read_single(each,points_per_request=1, discover_request=discover_request)
        else:
            self.read_single(points_list,points_per_request=1, discover_request=discover_request)
                        

    def read_single(self, request, *, points_per_request=1, discover_request=(None, 4)):
        try:
            request = ('{} {}'.format(self.properties.address, ''.join(request)))
            return self.properties.network.read(request)
        
        except KeyError as error:
            raise Exception('Unknown point name: %s' % error)
        
        except NoResponseFromController as error:
            return ''
            

    def _discoverPoints(self, custom_object_list = None):
        if custom_object_list:
            objList = custom_object_list
        else:
            try : 
                objList = self.properties.network.read('{} device {} objectList'.format(
                              self.properties.address, self.properties.device_id))
                
            except SegmentationNotSupported:
                objList = []
                number_of_objects = self.properties.network.read(
                    '{} device {} objectList'.format(self.properties.address, self.properties.device_id), arr_index = 0)
                
                for i in range(1,number_of_objects+1):
                    objList.append(self.properties.network.read(
                    '{} device {} objectList'.format(self.properties.address, self.properties.device_id), arr_index = i))

        points = []


        # Numeric
        for each in retrieve_type(objList, 'analog'):
            point_type = str(each[0])
            point_address = str(each[1])

            points.append(
                NumericPoint(
                    pointType=point_type,
                    pointAddress=point_address,
                    pointName=self.read_single('{} {} objectName '.format(point_type, point_address)),
                    description=self.read_single('{} {} description '.format(point_type, point_address)),
                    
                    presentValue=float(
                        self.read_single('{} {} presentValue '.format(point_type, point_address))),
                    units_state=self.read_single('{} {} units '.format(point_type, point_address)),
                    device=self))

        for each in retrieve_type(objList, 'multi'):
            point_type = str(each[0])
            point_address = str(each[1])

            points.append(
                EnumPoint(
                    pointType=point_type,
                    pointAddress=point_address,
                    pointName=self.read_single('{} {} objectName '.format(point_type, point_address)),
                    description=self.read_single('{} {} description '.format(point_type, point_address)),

                    presentValue=(self.read_single('{} {} presentValue '.format(point_type, point_address)),),
                    units_state=self.read_single('{} {} stateText '.format(point_type, point_address)),
                    device=self))

        for each in retrieve_type(objList, 'binary'):
            point_type = str(each[0])
            point_address = str(each[1])

            points.append(
                BooleanPoint(
                    pointType=point_type,
                    pointAddress=point_address,
                    pointName=self.read_single('{} {} objectName '.format(point_type, point_address)),
                    description=self.read_single('{} {} description '.format(point_type, point_address)),

                    presentValue=(self.read_single('{} {} presentValue '.format(point_type, point_address)),),
                    units_state=(
                                (self.read_single('{} {} inactiveText '.format(point_type, point_address))),
                                (self.read_single('{} {} activeText '.format(point_type, point_address)))
                                 ), 
                    device=self))

        self._log.info('Ready!')
        return (objList, points)

            
    def poll(self, command='start', *, delay=60):
        """
        Poll a point every x seconds (delay=x sec)
        Can be stopped by using point.poll('stop') or .poll(0) or .poll(False)
        or by setting a delay = 0

        :param command: (str) start or stop polling
        :param delay: (int) time delay between polls in seconds
        :type command: str
        :type delay: int

        :Example:

        device.poll()
        device.poll('stop')
        device.poll(delay = 5)
        """
        self._log.warning('Device too slow, use single points polling if needed')
        self._log.warning('Points will be read once...')
        for each in self.points:
            each.value
        self._log.info('Complete')



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
#
# Licensed under LGPLv3, see file LICENSE in this source tree.
"""
This will start the Bokeh Server
"""
from threading import Thread

from bokeh.server.server import Server

import weakref


class Bokeh_Worker(Thread):

    # Init thread running server
    def __init__(self, dev_app, trends_app, notes_app, serverIP, *, daemon=True):
        Thread.__init__(self, daemon=daemon)
        self._dev_app_ref = weakref.ref(dev_app)
        self._trends_app_ref = weakref.ref(trends_app)
        self._notes_app_ref = weakref.ref(notes_app)
        self.IP = serverIP
        self.exitFlag = False

    def run(self):
        self.process()

    def process(self):
        while not self.exitFlag:
            self.task()

    def startServer(self):
        self.server = Server({'/devices': self._dev_app_ref(),
                              '/trends': self._trends_app_ref(),
                              '/notes': self._notes_app_ref()},
                             allow_websocket_origin=[
                                 "%s:8111" % self.IP, "%s:5006" % self.IP],
                             address='0.0.0.0')
        self.server.start()
        self.server.io_loop.start()

    def task(self):
        try:
            self.startServer()
        except Exception as err:
            print('Bokeh server already running', err)
            self.exitFlag = True

    def stop(self):
        self.bokeh_server.stop()
        self.exitFlag = True

    def beforeStop(self):
        """
        Action done when closing thread
        """
        pass



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 13 21:37:02 2017

@author: CTremblay
"""


def create_sidebar(dash_class="", devices_class="", trends_class=""):
    sb = """
    	<div class="wrapper">
    <div class="sidebar" data-background-color="white" data-active-color="danger">

    <!--
		Tip 1: you can change the color of the sidebar's background using: data-background-color="white | black"
		Tip 2: you can change the color of the active button using the data-active-color="primary | info | success | warning | danger"
	-->

    	<div class="sidebar-wrapper">
            <div class="logo">
                <a href="https://github.com/ChristianTremblay/BAC0" class="simple-text">
                    BAC0
                </a>
            </div>

            <ul class="nav">
                <li %s>
                    <a href="/">
                        <i class="ti-panel"></i>
                        <p>Dashboard</p>
                    </a>
                </li>
                <li %s>
                    <a href="/trends">
                        <i class="ti-panel"></i>
                        <p>Trend</p>
                    </a>
                </li>
                <li %s>
                    <a href="/dash_devices">
                        <i class="ti-view-list-alt"></i>
                        <p>Devices</p>
                    </a>
                </li>
            </ul>
    	</div>
    </div>
    """ % (dash_class, trends_class, devices_class)
    return sb


def update_notifications(log, new_msg):
    notif_log = log
    max_notifications = 5
    if new_msg:
        notif_log.insert(0, new_msg)
    if len(notif_log) > max_notifications:
        notif_log.pop()

    notif_list = '<ul class="dropdown-menu">'
    for notif in notif_log:
        notif_list += '<li><a href="#">' + notif + "</a></li>"
    if len(notif_log) == 0:
        notif_list += '<li><a href="#">Nothing yet</a></li>'
    notif_list += '</ul>'
    return notif_list


def create_card(icon='ti-server', title='title',
                data='None', id_data='generic_data',
                foot_icon='ti-reload', foot_data='None',
                id_foot_data='generic_foot_data'):
    card = """
        <div class="col-lg-3 col-sm-6">
        <div class="card">
            <div class="content">
                <div class="row">
                    <div class="col-xs-5">
                        <div class="icon-big icon-success text-center">
                
            <i class="%s"></i>
                        </div>
                    </div>
                    <div class="col-xs-7">
                        <div class="numbers">
                            <p>%s</p>
                            <div id="%s">%s</div>
                        </div>
                    </div>
                </div>
                <div class="footer">
                    <hr />
                    <div class="stats">
                        <i class="%s"></i> <div id="%s">%s</div>
                    </div>
                </div>
            </div>
        </div>
        </div>
        """ % (icon, title, id_data, data, foot_icon, id_foot_data, foot_data)
    return card



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
discoverPoints.py - allow discovery of BACnet points in a controller.
'''

#--- standard Python modules ---

#--- 3rd party modules ---
try:
    import pandas as pd
    _PANDA = True
except ImportError:
    _PANDA = False

#--- this application's modules ---
from ..devices.Points import EnumPoint, BooleanPoint, NumericPoint


#------------------------------------------------------------------------------

def discoverPoints(bacnetapp, address, devID):
    """
    Discover the BACnet points in a BACnet device.

    :param bacnetApp: The app itself so we can call read
    :param address: address of the device as a string (ex. '2:5')
    :param devID: device ID of the bacnet device as a string (ex. '1001')

    :returns: a tuple with deviceName, pss, objList, df

        * *deviceName* : name of the device
        * *pss* : protocole service supported
        * *objList* : list of bacnet object (ex. analogInput, 1)
        * *df* : is a dataFrame containing pointType, pointAddress, pointName, description
        presentValue and units

        If pandas can't be found, df will be a simple array
    """
    pss = bacnetapp.read('{} device {} protocolServicesSupported'.format(address, devID))
    deviceName = bacnetapp.read('{} device {} objectName'.format(address, devID))

    print('Device {}- building points list'.format(deviceName))
    objList = bacnetapp.read('{} device {] objectList'.format(address, devID))

    newLine = []
    result = []
    points = []

    for pointType, pointAddr in objList:

        if 'binary' in pointType:           # BI/BO/BV
            newLine = [pointType, pointAddr]
            infos = bacnetapp.readMultiple(
                        '{} {} {} objectName description presentValue inactiveText activeText'.format(
                        address, pointType, pointAddr))

            newLine.extend(infos[:-2])
            newLine.extend([infos[-2:]])
            newPoint = BooleanPoint(pointType=newLine[0], pointAddress=newLine[1], 
                                    pointName=newLine[2], description=newLine[3], 
                                    presentValue=newLine[4], units_state=newLine[5])

        elif 'multiState' in pointType:     # MI/MV/MO
            newLine = [pointType, pointAddr]
            newLine.extend(bacnetapp.readMultiple(
                '{} {} {} objectName description presentValue stateText'.format(address, pointType, pointAddr)))
            
            newPoint = EnumPoint(pointType=newLine[0], pointAddress=newLine[1], 
                                 pointName=newLine[2], description=newLine[3], 
                                 presentValue=newLine[4], units_state=newLine[5])

        elif 'analog' in pointType:         # AI/AO/AV
            newLine = [pointType, pointAddr]
            newLine.extend(bacnetapp.readMultiple(
                '{} {} {} objectName description presentValue units'.format(address, pointType, pointAddr)))
            
            newPoint = NumericPoint(pointType=newLine[0], pointAddress=newLine[1], 
                                    pointName=newLine[2], description=newLine[3], 
                                    presentValue=newLine[4], units_state=newLine[5])

        else:
            continue        # skip

        result.append(newLine)
        points.append(newPoint)


    if _PANDA:
        df = pd.DataFrame(result, columns=['pointType', 'pointAddress', 'pointName',
                                           'description', 'presentValue', 'units_state']).set_index(['pointName'])
    else:
        df = result
        
    print('Ready!')
    return (deviceName, pss, objList, df, points)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
RecurringTask.py - execute a recurring task 
"""

from .TaskManager import Task

class RecurringTask(Task):
    """
    Start a recurring task (a function passed)
    """

    def __init__(self, fnc, delay = 60):
        """
        :param point: (BAC0.core.device.Points.Point) name of the point to read
        :param delay: (int) Delay between reads in seconds, defaults = 10sec
        
        A delay cannot be < 5sec (there are risks of overloading the device)

        :returns: Nothing
        """
        if hasattr(fnc, '__call__'):
            self.func = fnc
            Task.__init__(self)
        else:
            raise ValueError('You must pass a function to this...')

    def task(self):
        self.func()



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
infos.py - BAC0 Package MetaData
'''

__author__ = 'Christian Tremblay, P.Eng.'
__email__ = 'christian.tremblay@servisys.com'
__url__ = 'https://github.com/ChristianTremblay/BAC0'
__download_url__ = 'https://github.com/ChristianTremblay/BAC0/archive/master.zip'
__version__ = '0.99.920'
__license__ = 'LGPLv3'



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
SimpleApplication 
=================

A basic BACnet application (bacpypes BIPSimpleApplication) for interacting with 
the bacpypes BACnet stack.  It enables the base-level BACnet functionality 
(a.k.a. device discovery) - meaning it can send & receive WhoIs & IAm messages.

Additional functionality is enabled by inheriting this application, and then 
extending it with more functions. [See BAC0.scripts for more examples of this.]

'''
#--- standard Python modules ---
from collections import defaultdict

#--- 3rd party modules ---
from bacpypes.app import BIPSimpleApplication, BIPForeignApplication
from bacpypes.pdu import Address

#--- this application's modules ---
from ..utils.notes import note_and_log

#------------------------------------------------------------------------------


@note_and_log
class SimpleApplication(BIPSimpleApplication):
    """
    Defines a basic BACnet/IP application to process BACnet requests.

    :param *args: local object device, local IP address
        See BAC0.scripts.BasicScript for more details.

    """

    def __init__(self, *args, bbmdAddress=None, bbmdTTL=0):
        self.localAddress = None

        super().__init__(*args)

        self._request = None

        self.i_am_counter = defaultdict(int)
        self.who_is_counter = defaultdict(int)

        if isinstance(self.localAddress, Address):
            self.local_unicast_tuple = self.localAddress.addrTuple
            self.local_broadcast_tuple = self.localAddress.addrBroadcastTuple
        else:
            self.local_unicast_tuple = ('', 47808)
            self.local_broadcast_tuple = ('255.255.255.255', 47808)

    def do_WhoIsRequest(self, apdu):
        """Respond to a Who-Is request."""
        self.log(("do_WhoIsRequest {!r}".format(apdu)))

        # build a key from the source and parameters
        key = (str(apdu.pduSource),
               apdu.deviceInstanceRangeLowLimit,
               apdu.deviceInstanceRangeHighLimit)

        # count the times this has been received
        self.who_is_counter[key] += 1

        # continue with the default implementation
        BIPSimpleApplication.do_WhoIsRequest(self, apdu)

    def do_IAmRequest(self, apdu):
        """Given an I-Am request, cache it."""
        self.log(("do_IAmRequest {!r}".format(apdu)))

        # build a key from the source, just use the instance number
        key = (str(apdu.pduSource), apdu.iAmDeviceIdentifier[1])
        self.i_am_counter[key] += 1

        # continue with the default implementation
        BIPSimpleApplication.do_IAmRequest(self, apdu)


@note_and_log
class ForeignDeviceApplication(BIPForeignApplication):
    """
    Defines a basic BACnet/IP application to process BACnet requests.

    :param *args: local object device, local IP address
        See BAC0.scripts.BasicScript for more details.

    """

    def __init__(self, *args, bbmdAddress=None, bbmdTTL=0):
        self.localAddress = None

        super().__init__(*args, bbmdAddress=bbmdAddress, bbmdTTL=bbmdTTL)

        self._request = None

        self.i_am_counter = defaultdict(int)
        self.who_is_counter = defaultdict(int)

        if isinstance(self.localAddress, Address):
            self.local_unicast_tuple = self.localAddress.addrTuple
            self.local_broadcast_tuple = self.localAddress.addrBroadcastTuple
        else:
            self.local_unicast_tuple = ('', 47808)
            self.local_broadcast_tuple = ('255.255.255.255', 47808)

    def do_WhoIsRequest(self, apdu):
        """Respond to a Who-Is request."""
        self.log(("do_WhoIsRequest {!r}".format(apdu)))

        # build a key from the source and parameters
        key = (str(apdu.pduSource),
               apdu.deviceInstanceRangeLowLimit,
               apdu.deviceInstanceRangeHighLimit)

        # count the times this has been received
        self.who_is_counter[key] += 1

        # continue with the default implementation
        BIPSimpleApplication.do_WhoIsRequest(self, apdu)

    def do_IAmRequest(self, apdu):
        """Given an I-Am request, cache it."""
        self.log(("do_IAmRequest %r", apdu))

        # build a key from the source, just use the instance number
        key = (str(apdu.pduSource), apdu.iAmDeviceIdentifier[1])
        self.i_am_counter[key] += 1

        # continue with the default implementation
        BIPSimpleApplication.do_IAmRequest(self, apdu)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
Poll.py - create a Polling task to repeatedly read a point.
'''

#--- standard Python modules ---
import weakref

#--- 3rd party modules ---
from bacpypes.core import deferred

#--- this application's modules ---
from .TaskManager import Task

#------------------------------------------------------------------------------

class SimplePoll(Task):
    """
    Start a polling task to repeatedly read a point's Present_Value.
    ex.
        device['point_name'].poll(delay=60)
    """

    def __init__(self, point, *, delay=10):
        """
        :param point: (BAC0.core.device.Points.Point) name of the point to read
        :param delay: (int) Delay between reads in seconds, defaults = 10sec
        
        A delay cannot be < 5sec (there are risks of overloading the device)

        :returns: Nothing
        """
        if delay < 5:
            delay = 5
        if point.properties:
            self._point = point
            Task.__init__(self, name='rp_poll', delay=delay)
        else:
            raise ValueError('Provide a point object')

    def task(self):
        self._point.value


class DevicePoll(Task):
    """
    Start a polling task to repeatedly read a list of points from a device using 
    ReadPropertyMultiple requests.
    """

    def __init__(self, device, delay=10):
        """
        :param device: (BAC0.core.devices.Device.Device) device to poll
        :param delay: (int) Delay between polls in seconds, defaults = 10sec
        
        A delay cannot be < 5sec (there are risks of overloading the device)

        :returns: Nothing
        """
        if delay < 5:
            delay = 5
        self._device = weakref.ref(device)
        Task.__init__(self, name='rpm_poll', delay=delay, daemon = True)
        self._counter = 0

    @property
    def device(self):
        return self._device()

    def task(self):
        self.device.read_multiple(list(self.device.points_name), points_per_request=25)
        self._counter += 1
        if self._counter == self.device.properties.auto_save:
            self.device.save()
            if self.device.properties.clear_history_on_save:
                self.device.clear_histories()
            self._counter = 0



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
#
# Licensed under LGPLv3, see file LICENSE in this source tree.
"""
This module deals with Bokeh Session, Document and Plots
A connection to the server is mandatory to use update_data
"""
from bokeh.plotting import Figure
from bokeh.models import ColumnDataSource, HoverTool, Range1d, LinearAxis
from bokeh.models.widgets import DataTable, TableColumn, Div
from bokeh.layouts import widgetbox, row, column, gridplot
from bokeh.palettes import d3, Spectral6
from bokeh.io import curdoc
from bokeh.application.handlers import Handler

from tornado import gen

import math
import pandas as pd
import weakref

from ..tasks.RecurringTask import RecurringTask
from ..core.utils.notes import note_and_log


@note_and_log
class DynamicPlotHandler(Handler):
    def __init__(self, network):
        self._network = weakref.ref(network)
        self.sources = {}
        self._last_time_list = None
        self._update_complete = False
        self._recurring_update = RecurringTask(self.plan_update_data, delay=5)
        super().__init__()

    @property
    def network(self):
        return self._network()

    def organize_data(self):

        self._log.debug('Organize Data')

        self.s = {}
        for point in self.network.trends:
            self.s[point.history.name] = (point.history, point.history.units)
        self.lst_of_trends = [his[0] for name, his in self.s.items()]

    def build_plot(self):
        self._log.debug('Build Plot')

        self.stop_update_data()
        self.organize_data()
        for each in self.lst_of_trends:
            df = pd.DataFrame(each)
            df = df.reset_index()
            # print(df)
            df['name'] = each.name
            df['units'] = str(each.units)
            df['time_s'] = df['index'].apply(str)
            df.states = each.states
            try:
                df = df.fillna(method='ffill').fillna(
                    method='bfill').replace(['inactive', 'active'], [0, 1])
            except TypeError:
                df = df.fillna(method='ffill').fillna(method='bfill')

            self.sources[each.name] = ColumnDataSource(
                data=dict(
                    x=df['index'],
                    y=df[each.name],
                    time=df['time_s'],
                    name=df['name'],
                    units=df['units']
                )
            )

        TOOLS = "pan,box_zoom,wheel_zoom,save,reset"
        self.p = Figure(x_axis_type="datetime", x_axis_label="Time",
                        y_axis_label="Numeric Value", title='BAC0 Trends',
                        tools=TOOLS, plot_width=1300, plot_height=600,
                        toolbar_location='above')

        self.p.background_fill_color = "#f4f3ef"
        self.p.border_fill_color = "#f4f3ef"
        self.p.extra_y_ranges = {"bool": Range1d(start=0, end=1.1),
                                 "enum": Range1d(start=0, end=10)}
        self.p.add_layout(LinearAxis(y_range_name="bool",
                                     axis_label="Binary"), 'left')
        self.p.add_layout(LinearAxis(y_range_name="enum",
                                     axis_label="Enumerated"), 'right')

        hover = HoverTool(tooltips=[
            ('name', '@name'),
            ('value', '@y'),
            ('units', '@units'),
            ('time', '@time'),
        ])
        self.p.add_tools(hover)

        length = len(self.s.keys())
        if length <= 10:
            if length < 3:
                length = 3
            color_mapper = dict(zip(self.s.keys(), d3['Category10'][length]))
        else:
            # This would be a very loaded trend...
            color_mapper = dict(zip(self.s.keys(), Spectral6[:length]))

        for each in self.lst_of_trends:
            if each.states == 'binary':
                self.p.circle('x',
                              'y',
                              source=self.sources[each.name],
                              name=each.name,
                              color=color_mapper[each.name],
                              legend=("%s | %s (OFF-ON)" %
                                      (each.name, each.description)),
                              y_range_name="bool",
                              size=10)
            elif each.states == 'multistates':
                self.p.diamond('x',
                               'y',
                               source=self.sources[each.name],
                               name=each.name,
                               color=color_mapper[each.name],
                               legend=("%s | %s (%s)" %
                                       (each.name, each.description, each.units)),
                               y_range_name="enum",
                               size=20)
            else:
                self.p.line('x',
                            'y',
                            source=self.sources[each.name],
                            name=each.name,
                            color=color_mapper[each.name],
                            legend=("%s | %s (%s)" %
                                    (each.name, each.description, each.units)),
                            line_width=2)

            self.p.legend.location = 'bottom_left'
            self.p.legend.click_policy = "hide"

            self.plots = [self.p, ]

    #@gen.coroutine
    def update_data(self):
        self._log.debug('Update Data')
        doc = curdoc()
        self.organize_data()
        if self._last_time_list:
            if self._last_time_list != self.s.keys():
                self._list_have_changed = True
                self.stop_update_data()
                self.modify_document(doc)
            else:
                self._list_have_changed = False

        l = []
        for each in self.p.renderers:
            l.append(each.name)

        for each in self.lst_of_trends:
            df = pd.DataFrame(each)
            df = df.reset_index()
            df['name'] = each.name
            df['units'] = str(each.units)
            df['time_s'] = df['index'].apply(str)

            try:
                df = df.fillna(method='ffill').fillna(
                    method='bfill').replace(['inactive', 'active'], [0, 1])
            except TypeError:
                df = df.fillna(method='ffill').fillna(method='bfill')

            index = l.index(each.name)
            renderer = self.p.renderers[index]
            new_data = {}
            new_data['name'] = df['name']
            new_data['x'] = df['index']
            new_data['y'] = df[each.name]
            if each.states == 'binary':
                new_data['units'] = [each.units[int(x)] for x in df[each.name]]
            elif each.states == 'multistates':
                new_data['units'] = [
                    each.units[int(math.fabs(x-1))] for x in df[each.name]]
            else:
                new_data['units'] = df['units']
            new_data['time'] = df['time_s']
            renderer.data_source.data = new_data
        self._last_time_list = self.s.keys()
        # self.start_update_data()
        self._update_complete = True

    def modify_document(self, doc):
        curdoc().clear()
        try:
            curdoc().remove_periodic_callback(self.update_data)
        except:
            pass
        doc.clear()
        self.build_plot()
        # column(widgetbox(self.div_header_doc),
        layout = gridplot(self.plots, ncols=2)
        # widgetbox(self.div_header_notes),
        # widgetbox(self.div_footer)

        doc.add_root(layout)
        doc.add_periodic_callback(self.update_data, 10000)
        return doc

    def plan_update_data(self):
        doc = curdoc()
        if self._update_complete == True:
            self._update_complete = False
            doc.add_next_tick_callback(self.update_data)

    def stop_update_data(self):
        doc = curdoc()
        if self._recurring_update.is_running:
            self._recurring_update.stop()
            while self._recurring_update.is_running:
                pass
            try:
                doc.remove_next_tick_callback(self.update_data)
            except (ValueError, RuntimeError):
                pass  # Already gone

    def start_update_data(self):
        if not self._recurring_update.is_running:
            try:
                self._recurring_update.start()
                while not self._recurring_update.is_running:
                    pass
            except RuntimeError:
                pass


class DevicesTableHandler(Handler):
    """ 
    This handler will poll the network and show devices.

    """

    def __init__(self, network):
        self._network = weakref.ref(network)
        super().__init__()

    @property
    def network(self):
        return self._network()

    def modify_document(self, doc):
        self.network.whois()
        devices_df = self.network.devices
        dev = ColumnDataSource(devices_df)
        columns = [
            TableColumn(field=" Device ID", title="Dev ID"),
            TableColumn(field="Address", title="Address"),
            TableColumn(field="Manufacturer", title="Manuf"),
            TableColumn(field="Name", title="Name")]
        data_table = DataTable(source=dev, columns=columns)
        layout = row([data_table])
        doc.add_root(layout)
        doc.title = 'BACnet devices'
        return doc


class NetworkPieChartHandler(Handler):
    """ 
    This handler will poll the network and show devices.

    """

    def __init__(self, network):
        self._network = weakref.ref(network)
        super().__init__()

    @property
    def network(self):
        return self._network()

    def modify_document(self, doc):
        self.network.whois()
        devices_df = self.network.devices
        dev = ColumnDataSource(devices_df)
        columns = [
            TableColumn(field=" Device ID", title="Dev ID"),
            TableColumn(field="Address", title="Address"),
            TableColumn(field="Manufacturer", title="Manuf"),
            TableColumn(field="Name", title="Name")]
        data_table = DataTable(source=dev, columns=columns)
        layout = row([data_table])
        doc.add_root(layout)
        doc.title = 'BACnet devices'
        return doc


class NotesTableHandler(Handler):
    """ 
    This handler will poll the network and show devices.

    """

    def __init__(self, network):
        self._network_ref = weakref.ref(network)
        super().__init__()

    @property
    def network(self):
        return self._network_ref()

    def modify_document(self, doc):

        controller = self.network.notes[0]
        notes_df = pd.DataFrame(self.network.notes[1]).reset_index()
        notes_df.columns = ['index', 'notes']
        notes = ColumnDataSource(notes_df)
        self.columns = [
            TableColumn(field="index", title="Timestamp"),
            TableColumn(field="notes", title="Notes")]
        self.data_table = DataTable(source=notes, columns=self.columns)
        layout = row([self.data_table])
        doc.add_root(layout)
        doc.title = 'Notes for %s' % controller
        # doc.add_periodic_callback(self.update_data,100)
        return doc

    def update_data(self):
        controller = self.network.notes[0]
        notes_df = pd.DataFrame(self.network.notes[1]).reset_index()
        notes_df.columns = ['index', 'notes']
        notes = ColumnDataSource(notes_df)
        self.data_table.source.data.update(notes.data)
        curdoc().title = 'Notes for %s' % controller



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
"""
BasicScript - implement the BAC0.core.app.ScriptApplication
Its basic function is to start and stop the bacpypes stack.
Stopping the stack, frees the IP socket used for BACnet communications. 
No communications will occur if the stack is stopped.

Bacpypes stack enables Whois and Iam functions, since this minimum is needed to be 
a BACnet device.  Other stack services can be enabled later (via class inheritance).
[see: see BAC0.scripts.ReadWriteScript]

Class::
    BasicScript(WhoisIAm)
        def startApp()
        def stopApp()
"""
#--- standard Python modules ---
from threading import Thread
import random
import sys

#--- 3rd party modules ---

from bacpypes.core import run as startBacnetIPApp
from bacpypes.core import stop as stopBacnetIPApp
from bacpypes.core import enable_sleeping
from bacpypes.local.device import LocalDeviceObject
from bacpypes.basetypes import ServicesSupported, DeviceStatus
from bacpypes.primitivedata import CharacterString

#--- this application's modules ---
from ..core.app.ScriptApplication import SimpleApplication, ForeignDeviceApplication
from .. import infos

from ..core.utils.notes import note_and_log

#------------------------------------------------------------------------------


@note_and_log
class Base():
    """
    Build a running BACnet/IP device that accepts WhoIs and IAm requests
    Initialization requires some minimial information about the local device.

    :param localIPAddr='127.0.0.1':
    :param localObjName='BAC0':
    :param DeviceId=None:
    :param maxAPDULengthAccepted='1024':
    :param maxSegmentsAccepted='1024':
    :param segmentationSupported='segmentedBoth':
    """

    def __init__(self, localIPAddr='127.0.0.1', localObjName='BAC0', DeviceId=None,
                 maxAPDULengthAccepted='1024', maxSegmentsAccepted='1024',
                 segmentationSupported='segmentedBoth',
                 bbmdAddress=None, bbmdTTL=0):

        self._log.debug("Configurating app")

        self.response = None
        self._initialized = False
        self._started = False
        self._stopped = False

        self.localIPAddr = localIPAddr
        self.Boid = int(DeviceId) if DeviceId else (
            3056177 + int(random.uniform(0, 1000)))

        self.segmentationSupported = segmentationSupported
        self.maxSegmentsAccepted = maxSegmentsAccepted
        self.localObjName = localObjName

        self.maxAPDULengthAccepted = maxAPDULengthAccepted
        self.vendorId = 842
        self.vendorName = CharacterString('SERVISYS inc.')
        self.modelName = CharacterString('BAC0 Scripting Tool')

        self.discoveredDevices = None
        self.systemStatus = DeviceStatus(1)

        self.bbmdAddress = bbmdAddress
        self.bbmdTTL = bbmdTTL

        self.startApp()

    def startApp(self):
        """
        Define the local device, including services supported.
        Once defined, start the BACnet stack in its own thread.
        """
        self._log.debug("Create Local Device")
        try:
            # make a device object
            self.this_device = LocalDeviceObject(
                objectName=self.localObjName,
                objectIdentifier=self.Boid,
                maxApduLengthAccepted=int(self.maxAPDULengthAccepted),
                segmentationSupported=self.segmentationSupported,
                vendorIdentifier=self.vendorId,
                vendorName=self.vendorName,
                modelName=self.modelName,
                systemStatus=self.systemStatus,
                description='http://christiantremblay.github.io/BAC0/',
                firmwareRevision=''.join(sys.version.split('|')[:2]),
                applicationSoftwareVersion=infos.__version__,
                protocolVersion=1,
                protocolRevision=0,
            )

            # build a bit string that knows about the bit names
#            pss = ServicesSupported()
#            pss['whoIs'] = 1
#            pss['iAm'] = 1
#            pss['readProperty'] = 1
#            pss['writeProperty'] = 1
#            pss['readPropertyMultiple'] = 1
#
#            # set the property value to be just the bits
#            self.this_device.protocolServicesSupported = pss.value

            # make an application
            if self.bbmdAddress and self.bbmdTTL > 0:

                self.this_application = ForeignDeviceApplication(self.this_device, self.localIPAddr,
                                                                 bbmdAddress=self.bbmdAddress, bbmdTTL=self.bbmdTTL)
                app_type = 'Foreign Device'
            else:
                self.this_application = SimpleApplication(
                    self.this_device, self.localIPAddr)
                app_type = 'Simple BACnet/IP App'
            self._log.debug("Starting")
            self._initialized = True
            try:
                self._startAppThread()
                print("Registered as {}".format(app_type))
            except:
                self._log.warning("Error opening socket")
                raise
            self._log.debug("Running")

        except Exception as error:
            self._log.error("an error has occurred: {}".format(error))
        finally:
            self._log.debug("finally")

    def register_foreign_device(self, addr=None, ttl=0):
        self.this_application.bip.register(addr, ttl)

    def unregister_foreign_device(self):
        self.this_application.bip.unregister()

    def disconnect(self):
        """
        Stop the BACnet stack.  Free the IP socket.
        """
        print('Stopping BACnet stack')
        # Freeing socket
        try:
            self.this_application.mux.directPort.handle_close()
        except:
            self.this_application.mux.broadcastPort.handle_close()

        stopBacnetIPApp()           # Stop Core
        self._stopped = True        # Stop stack thread
        self.t.join()
        self._started = False
        print('BACnet stopped')

    def _startAppThread(self):
        """
        Starts the BACnet stack in its own thread so requests can be processed.
        As signal cannot be called in another thread than the main thread
        when calling startBacnetIPApp, we must pass None to both parameters
        """
        self._log.info('Starting app...')
        enable_sleeping(0.0005)
        self.t = Thread(target=startBacnetIPApp, kwargs={
                        'sigterm': None, 'sigusr1': None}, daemon=True)
        self.t.start()
        self._started = True
        self._log.info('BAC0 started')



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
#
# Licensed under LGPLv3, see file LICENSE in this source tree.
"""
This module will start the Flask server 
"""
from threading import Thread
import weakref

from flask import Flask, render_template, jsonify, request
import json
from bokeh.embed import server_document

from .templates import create_sidebar, create_card, update_notifications


class FlaskServer(Thread):

    # Init thread running server
    def __init__(self, network, port=8111, ip='0.0.0.0', *, daemon=True):
        Thread.__init__(self, daemon=daemon)
        self.flask_app = Flask(__name__)
        self._network_ref = weakref.ref(network)
        self.port = port
        self.ip = ip
        self.notifications_log = []
        self.notifications_list = ""
        self.config_flask_app()
        self.exitFlag = False

    @property
    def network(self):
        return self._network_ref()

    def run(self):
        self.process()

    def process(self):
        while not self.exitFlag:
            self.task()

    def startServer(self):
        self.flask_app.run(port=self.port, host='0.0.0.0')

    def config_flask_app(self):
        @self.flask_app.route('/trends', methods=['GET'])
        def bkapp_trends_page():
            if self.network.number_of_registered_trends > 0:
                script = server_document('http://%s:5006/trends' % self.ip)
            else:
                script = "<div>No trend registered yet...</div>"
            return render_template("trends.html",
                                   sidebar=create_sidebar(
                                       trends_class='class="active"'),
                                   bokeh_script=script,
                                   template="Flask")

#        @self.flask_app.route('/devices', methods=['GET'])
#        def bkapp_devices_page():
#            script = server_document('http://localhost:5006/devices')
#            return render_template("embed.html", script=script, template="Flask")

        @self.flask_app.route('/notes', methods=['GET'])
        def bkapp_notes_page():
            script = server_document('http://%s:5006/notes' % self.ip)
            return render_template("embed.html", script=script, template="Flask")

        @self.flask_app.route('/', methods=['GET'])
        def home_page():
            # Stat number of devices
            cnod = create_card(icon='ti-server',
                               title='Number of devices',
                               data=self.network.number_of_devices,
                               id_data='devices',
                               foot_icon='ti-reload',
                               foot_data='Refresh to update')
            cnot = create_card(icon='ti-bar-chart',
                               title='Number of trends',
                               data=self.network.number_of_registered_trends,
                               id_data='trends',
                               foot_icon='ti-timer',
                               foot_data='Add trends !')

            cnmn = create_card(icon='ti-plug',
                               title='%s MSTP Networks' % len(
                                   self.network.network_stats['mstp_networks']),
                               data='# %s' % (
                                   self.network.network_stats['print_mstpnetworks']),
                               id_data='mstpnetworks',
                               foot_icon='ti-timer',
                               foot_data='Last update : %s' % self.network.network_stats['timestamp'],
                               id_foot_data='lastwhoisupdate')

            notif = update_notifications(
                self.notifications_log, None)
            return render_template("dashboard.html",
                                   sidebar=create_sidebar(
                                       dash_class='class="active"'),
                                   card_number_of_devices=cnod,
                                   card_number_of_mstp_networks=cnmn,
                                   card_number_of_trends=cnot,
                                   notifications=notif,
                                   template="Flask")

        @self.flask_app.route('/dash_devices', methods=['GET'])
        def dashboard_devices_page():
            script = server_document('http://%s:5006/devices' % self.ip)
            return render_template("device_table.html",
                                   sidebar=create_sidebar(
                                       devices_class='class="active"'),
                                   bokeh_script=script,
                                   template="Flask")

        @self.flask_app.route('/_dash_live_data', methods=['GET'])
        def dash_live_data():
            devices = self.network.number_of_devices
            trends = self.network.number_of_registered_trends
            return jsonify(number_of_devices=devices, number_of_registered_trends=trends)

        @self.flask_app.route('/_whois', methods=['GET'])
        def whois():
            self.notifications_list = update_notifications(
                self.notifications_log, 'Sent a WhoIs Request')
            self.network.whois_answer = self.network.update_whois()
            return jsonify(done='done')

        @self.flask_app.route('/_dash_live_stats', methods=['GET'])
        def dash_live_stats():
            stats = self.network.network_stats
            return jsonify(stats=stats)

        @self.flask_app.route('/_network_pie_chart', methods=['GET'])
        def net_pie_chart():
            stats = self.network.number_of_devices_per_network()
            return jsonify(stats=stats)

        @self.flask_app.route('/log', methods=['POST', 'GET'])
        def log_page():
            return json.dumps(request.form)

    def task(self):
        try:
            self.startServer()
        except Exception as err:
            print('Flask server already running', err)
            self.exitFlag = True

    def stop(self):
        print('Trying to stop Bokeh Server')
        # self.bokeh_server.stop()
        self.p.terminate()
        self.exitFlag = True

    def beforeStop(self):
        """
        Action done when closing thread
        """
        pass



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
Match.py - verify a point's status matches its commanded value.

Example:
    Is a fan commanded to 'On' actually 'running'?
'''

#--- standard Python modules ---
#--- 3rd party modules ---

#--- this application's modules ---
from .TaskManager import Task

#------------------------------------------------------------------------------

class Match(Task):
    """
    Match two properties of a BACnet Object (i.e. a point status with its command).
    """

    def __init__(self, command = None, status = None, delay=5):        
        self.command = command
        self.status = status
        Task.__init__(self, delay=delay, daemon = True)


    def task(self):
        if  self.status.history[-1] != self.command.history[-1]:
            self.status._setitem(self.command.history[-1])        


    def stop(self):
        self.status._setitem('auto')
        self.exitFlag = True


class Match_Value(Task):
    """
    Verify a point's Present_Value equals the given value after a delay of X seconds.
    Thus giving the BACnet controller (and connected equipment) time to respond to the
    command.
    
        Match_Value(On, <AI:1>, 5)
        
    i.e. Does Fan value = On after 5 seconds. 
    """

    def __init__(self, value = None, point = None, delay=5):        
        self.value = value
        self.point = point
        Task.__init__(self, delay=delay, daemon = True)


    def task(self):
        if hasattr(self.value, '__call__'):
            value = self.value()
        else:
            value = self.value
        if  value != self.point:
            self.point._setitem(value)        


    def stop(self):
        self.status._setitem('auto')
        self.exitFlag = True



# -*- coding: utf-8 -*-
"""
Notes and logger decorator to be used on class

This will add a "notes" object to the class and will allow
logging feature at the same time.
Goal is to be able to access quickly to important informations for
the web interface.

"""
#--- standard Python modules ---
from collections import namedtuple
from datetime import datetime
import logging
from logging import FileHandler
import sys

import os
from os.path import expanduser, join

#--- 3rd party modules ---
try:
    import pandas as pd
    _PANDAS = True
except ImportError:
    _PANDAS = False

def convert_level(level):
    if not level:
        return None
    if level.lower() == 'info':
        level = logging.INFO
    elif level.lower() == 'debug':
        level = logging.DEBUG
    elif level.lower() == 'warning':
        level = logging.WARNING
    elif level.lower() == 'error':
        level = logging.ERROR
    elif level.lower() == 'critical':
        level = logging.CRITICAL
    return level

def update_log_level(file, stderr = None, stdout = None):
    """
    Typical usage : 
        Normal
        BAC0.log_level(file='warning', stdout='warning', stderr='error')
        Info on console....but not in file
        BAC0.log_level(file='warning', stdout='info', stderr='error')
        Debug
        BAC0.log_level(file='debug', stdout='info', stderr='error')
    """
    file = convert_level(file)
    stderr = convert_level(stderr)
    stdout = convert_level(stdout)       
    BAC0_logger = logging.getLogger('BAC0')
#    if console:
#        BAC0_logger.setLevel(console)
#        BAC0_logger.warning('Changed log level of console to {}'.format(logging.getLevelName(level)))

    for handler in BAC0_logger.handlers:
        if file and handler.get_name() == 'file_handler':
            handler.setLevel(file)
            BAC0_logger.info('Changed log level of file to {}'.format(logging.getLevelName(file)))
        elif stdout and handler.get_name() == 'stdout':
            handler.setLevel(stdout)
            BAC0_logger.info('Changed log level of console stdout to {}'.format(logging.getLevelName(stdout)))
        elif stderr and handler.get_name() == 'stderr':
            handler.setLevel(stderr)
            BAC0_logger.info('Changed log level of console stderr to {}'.format(logging.getLevelName(stderr)))


def note_and_log(cls):
    """
    This will be used as a decorator on class to activate
    logging and store messages in the variable cls._notes
    This will allow quick access to events in the web app.

    A note can be added to cls._notes without logging if passing
    the argument log=false to function note()
    Something can be logged without addind a note using function log()
    """
    if hasattr(cls, 'DEBUG_LEVEL'):
        if cls.DEBUG_LEVEL == 'debug':
            file_level = logging.DEBUG
            console_level = logging.DEBUG
        elif cls.DEBUG_LEVEL == 'info':
            file_level = logging.INFO
            console_level = logging.INFO
    else:
        file_level = logging.INFO
        console_level = logging.WARNING
    # Notes object
    cls._notes = namedtuple('_notes', ['timestamp', 'notes'])
    cls._notes.timestamp = []
    cls._notes.notes = []

    # Defining log object
    cls.logname = '{} | {}'.format(cls.__module__, cls.__name__)
    root_logger = logging.getLogger()
    cls._log = logging.getLogger('BAC0')
    if not len(root_logger.handlers):
        root_logger.addHandler(cls._log)
    
    # Console Handler
    ch = logging.StreamHandler()
    ch.set_name('stderr')
    ch2 = logging.StreamHandler(sys.stdout)
    ch2.set_name('stdout')
    ch.setLevel(console_level)
    ch2.setLevel(logging.CRITICAL)

    formatter = logging.Formatter(
        '{asctime} - {levelname:<8}| {message}', style='{')

    # File Handler
    _PERMISSION_TO_WRITE = True
    logUserPath = expanduser('~')
    logSaveFilePath = join(logUserPath, '.BAC0')

    logFile = join(logSaveFilePath, 'BAC0.log')
    if not os.path.exists(logSaveFilePath):
        try:
            os.makedirs(logSaveFilePath)
        except:
            _PERMISSION_TO_WRITE = False
    if _PERMISSION_TO_WRITE:
        fh = FileHandler(logFile)
        fh.set_name('file_handler')
        fh.setLevel(file_level)
        fh.setFormatter(formatter)

    ch.setFormatter(formatter)
    ch2.setFormatter(formatter)
    # Add handlers the first time only...
    if not len(cls._log.handlers):
        if _PERMISSION_TO_WRITE:
            cls._log.addHandler(fh)
        cls._log.addHandler(ch)
        cls._log.addHandler(ch2)
    
#    cls._log.setLevel(logging.CRITICAL)
        
    def log_title(self, title, args=None, width=35):
        cls._log.info("")
        cls._log.info("#"*width)
        cls._log.info("# {}".format(title))
        cls._log.info("#"*width)
        if args:
            cls._log.debug("{!r}".format(args))
            cls._log.debug("#"*35)

    def log_subtitle(self, subtitle, args=None, width=35):
        cls._log.info("")
        cls._log.info("="*width)
        cls._log.info("{}".format(subtitle))
        cls._log.info("="*width)
        if args:
            cls._log.debug("{!r}".format(args))
            cls._log.debug("="*width)

    def log(self, note, *, level=logging.DEBUG):
        """
        Add a log entry...no note
        """
        if not note:
            raise ValueError('Provide something to log')
        note = '{} | {}'.format(cls.logname, note)
        cls._log.log(level, note)

    def note(self, note, *, level=logging.INFO, log=True):
        """
        Add note to the object. By default, the note will also
        be logged

        :param note: (str) The note itself
        :param level: (logging.level)
        :param log: (boolean) Enable or disable logging of note
        """
        if not note:
            raise ValueError('Provide something to log')
        note = '{} | {}'.format(cls.logname, note)
        cls._notes.timestamp.append(datetime.now())
        cls._notes.notes.append(note)
        if log:
            cls.log(level, note)

    @property
    def notes(self):
        """
        Retrieve notes list as a Pandas Series
        """
        if not _PANDAS:
            return dict(zip(self._notes.timestamp, self._notes.notes))
        return pd.Series(self._notes.notes, index=self._notes.timestamp)

    def clear_notes(self):
        """
        Clear notes object
        """
        cls._notes.timestamp = []
        cls._notes.notes = []

    # Add the functions to the decorated class
    cls.clear_notes = clear_notes
    cls.note = note
    cls.notes = notes
    cls.log = log
    cls.log_title = log_title
    cls.log_subtitle = log_subtitle
    return cls



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
ReadWriteScript - extended version of BasicScript.py

As everything is handled by the BasicScript, select the additional features you want::

    # Create a class that implements a basic script with read and write functions
    from BAC0.scripts.BasicScript import BasicScript
    from BAC0.core.io.Read import ReadProperty
    from BAC0.core.io.Write import WriteProperty
    class ReadWriteScript(BasicScript,ReadProperty,WriteProperty)

Once the class is created, create the local object and use it::

    bacnet = ReadWriteScript(localIPAddr = '192.168.1.10')
    bacnet.read('2:5 analogInput 1 presentValue)

'''
#--- standard Python modules ---
import time
from datetime import datetime
import weakref


#--- this application's modules ---
from ..scripts.Base import Base

from ..core.io.Read import ReadProperty
from ..core.io.Write import WriteProperty
from ..core.functions.GetIPAddr import HostIP
from ..core.functions.WhoisIAm import WhoisIAm
from ..core.io.Simulate import Simulation
from ..core.devices.Points import Point
from ..core.utils.notes import note_and_log
from ..core.io.IOExceptions import NoResponseFromController, UnrecognizedService

from ..infos import __version__ as version


#------------------------------------------------------------------------------


@note_and_log
class Lite(Base, WhoisIAm, ReadProperty, WriteProperty, Simulation):
    """
    Build a BACnet application to accept read and write requests.
    [Basic Whois/IAm functions are implemented in parent BasicScript class.]
    Once created, execute a whois() to build a list of available controllers.
    Initialization requires information on the local device.

    :param ip='127.0.0.1': Address must be in the same subnet as the BACnet network
        [BBMD and Foreign Device - not supported]

    :param bokeh_server: (boolean) If set to false, will prevent Bokeh server
        from being started. Can help troubleshoot issues with Bokeh. By default,
        set to True.
    """

    def __init__(self, ip=None, bbmdAddress=None, bbmdTTL=0):
        print("Starting BAC0 version {} ({})".format(
              version, self.__module__.split('.')[-1]))
        self.log("Configurating app")
        self._registered_devices = weakref.WeakValueDictionary()
        if ip is None:
            host = HostIP()
            ip_addr = host.address
        else:
            ip_addr = ip

        Base.__init__(self, localIPAddr=ip_addr,
                      bbmdAddress=bbmdAddress, bbmdTTL=bbmdTTL)

        self.bokehserver = False
        self._points_to_trend = weakref.WeakValueDictionary()
        # Force a global whois to find all devices on the network
        self.whois_answer = self.update_whois()
        time.sleep(2)

    def update_whois(self):
        return (self.whois(), str(datetime.now()))

    def register_device(self, device):
        oid = id(device)
        self._registered_devices[oid] = device

    @property
    def registered_devices(self):
        """
        Devices that have been created using BAC0.device(args)
        """
        return list(self._registered_devices.values())

    def unregister_device(self, device):
        """
        Remove from the registered list
        """
        oid = id(device)
        del self._registered_devices[oid]

    def add_trend(self, point_to_trend):
        """
        Add point to the list of histories that will be handled by Bokeh

        Argument provided must be of type Point
        ex. bacnet.add_trend(controller['point_name'])
        """
        if isinstance(point_to_trend, Point):
            oid = id(point_to_trend)
            self._points_to_trend[oid] = point_to_trend
        else:
            raise TypeError('Please provide point containing history')

    def remove_trend(self, point_to_remove):
        """
        Remove point from the list of histories that will be handled by Bokeh

        Argument provided must be of type Point
        ex. bacnet.remove_trend(controller['point_name'])
        """
        if isinstance(point_to_remove, Point):
            oid = id(point_to_remove)
        else:
            raise TypeError('Please provide point containing history')
        if oid in self._points_to_trend.keys():
            del self._points_to_trend[oid]

    @property
    def devices(self):
        lst = []
        for device in list(self.discoveredDevices):
            try:
                deviceName, vendorName = self.readMultiple(
                    '{} device {} objectName vendorName'.format(device[0], device[1]))
            except UnrecognizedService:
                deviceName = self.read(
                    '{} device {} objectName'.format((device[0], device[1])))
                vendorName = self.read(
                    '{} device {} vendorName'.format(device[0], device[1]))
            except NoResponseFromController:
                self._log.info('No response from {}'.format(device))
                continue
            lst.append((deviceName, vendorName, device[0], device[1]))
        return lst

    @property
    def trends(self):
        """
        This will present a list of all registered trends used by Bokeh Server
        """
        return list(self._points_to_trend.values())

    def disconnect(self):
        super().disconnect()

    def __repr__(self):
        return 'Bacnet Network using ip %s with device id %s' % (self.localIPAddr, self.Boid)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
Write.py - creation of WriteProperty requests

    Used while defining an app
    Example::

        class BasicScript(WhoisIAm, WriteProperty)

    Class::

        WriteProperty()
            def write()

    Functions::

        print_debug()

'''
#--- 3rd party modules ---
from bacpypes.debugging import bacpypes_debugging, ModuleLogger

from bacpypes.pdu import Address
from bacpypes.object import get_datatype

from bacpypes.apdu import WritePropertyRequest, SimpleAckPDU

from bacpypes.primitivedata import Null, Atomic, Integer, Unsigned, Real
from bacpypes.constructeddata import Array, Any
from bacpypes.iocb import IOCB
from bacpypes.core import deferred

#--- this application's modules ---
from .IOExceptions import WritePropertyCastError, NoResponseFromController, WritePropertyException, WriteAccessDenied, ApplicationNotStarted
from ...core.utils.notes import note_and_log

#------------------------------------------------------------------------------

# some debugging
_debug = 0
_LOG = ModuleLogger(globals())


@note_and_log
class WriteProperty():
    """
    Defines BACnet Write functions: WriteProperty [WritePropertyMultiple not supported]

    """

    def write(self, args, vendor_id=0):
        """ Build a WriteProperty request, wait for an answer, and return status [True if ok, False if not].

        :param args: String with <addr> <type> <inst> <prop> <value> [ <indx> ] [ <priority> ]
        :returns: data read from device (str representing data like 10 or True)

        *Example*::

            import BAC0
            myIPAddr = '192.168.1.10'
            bacnet = BAC0.ReadWriteScript(localIPAddr = myIPAddr)
            bacnet.write('2:5 analogValue 1 presentValue 100')

        Direct the controller at (Network 2, address 5) to write 100 to the presentValues of
        its analogValue 1 (AV:1)
        """
        if not self._started:
            raise ApplicationNotStarted(
                'BACnet stack not running - use startApp()')
        args = args.split()
        self.log_title("Write property",args)

        try:
            # build a WriteProperty request
            iocb = IOCB(self.build_wp_request(args, vendor_id=vendor_id))
            # pass to the BACnet stack
            deferred(self.this_application.request_io, iocb)
            self._log.debug("{:<20} {!r}".format('iocb', iocb))

        except WritePropertyException as error:
            # construction error
            self._log.exception("exception: {!r}".format(error))

        iocb.wait()             # Wait for BACnet response

        if iocb.ioResponse:     # successful response
            apdu = iocb.ioResponse
            
            if not isinstance(iocb.ioResponse, SimpleAckPDU):   # expect an ACK
                self._log.warning("Not an ack, see debug for more infos.")
                self._log.debug("Not an ack. | APDU : {} / {}".format((apdu, type(apdu))))
                return

        if iocb.ioError:        # unsuccessful: error/reject/abort
            raise NoResponseFromController()

    def build_wp_request(self, args, vendor_id=0):
        addr, obj_type, obj_inst, prop_id = args[:4]
        vendor_id = vendor_id
        if obj_type.isdigit():
            obj_type = int(obj_type)
        obj_inst = int(obj_inst)
        value = args[4]

        indx = None
        if len(args) >= 6:
            if args[5] != "-":
                indx = int(args[5])

        priority = None
        if len(args) >= 7:
            priority = int(args[6])

        # get the datatype
        if prop_id.isdigit():
            prop_id = int(prop_id)
        datatype = get_datatype(obj_type, prop_id, vendor_id=vendor_id)

        self.log_subtitle("Creating Request")
        self._log.info(
                "{:<20} {:<20} {:<20} {:<20}".format(
                'indx',
                'priority',
                'datatype',
                'value'))
        self._log.info(
                "{!r:<20} {!r:<20} {!r:<20} {!r:<20}".format(
                indx,
                priority,
                datatype,
                value
                ))
        
        # change atomic values into something encodeable, null is a special
        # case
        if value == 'null':
            value = Null()

        elif issubclass(datatype, Atomic):
            if datatype is Integer:
                value = int(value)
            elif datatype is Real:
                value = float(value)
            elif datatype is Unsigned:
                value = int(value)
            value = datatype(value)

        elif issubclass(datatype, Array) and (indx is not None):
            if indx == 0:
                value = Integer(value)
            elif issubclass(datatype.subtype, Atomic):
                value = datatype.subtype(value)
            elif not isinstance(value, datatype.subtype):
                raise TypeError(
                    "invalid result datatype, expecting {}".format(
                        (datatype.subtype.__name__,)))

        elif not isinstance(value, datatype):
            raise TypeError(
                "invalid result datatype, expecting {}".format(
                    (datatype.__name__,)))
        self._log.info("{:<20} {!r} {}".format(
            "Encodeable value", value, type(value)))

        # build a request
        request = WritePropertyRequest(objectIdentifier=(obj_type, obj_inst),
                                       propertyIdentifier=prop_id)
        request.pduDestination = Address(addr)

        # save the value
        request.propertyValue = Any()
        try:
            request.propertyValue.cast_in(value)
        except WritePropertyCastError as error:
            self._log.error("WriteProperty cast error: {!r}".format(error))

        # optional array index
        if indx is not None:
            request.propertyArrayIndex = indx

        # optional priority
        if priority is not None:
            request.priority = priority

        self._log.debug("{:<20} {}".format("REQUEST", request))
        return request



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
WhoisIAm.py - creation of Whois and IAm requests

Used while defining an app
ex.: class BasicScript(WhoisIAm):

Class : WhoisIAm

'''
#--- standard Python modules ---

#--- 3rd party modules ---
from bacpypes.debugging import bacpypes_debugging
from bacpypes.apdu import WhoIsRequest, IAmRequest

from bacpypes.pdu import Address, GlobalBroadcast
from bacpypes.primitivedata import Unsigned
from bacpypes.constructeddata import Array
from bacpypes.object import get_object_class, get_datatype
from bacpypes.iocb import IOCB

#--- this application's modules ---
from ..io.IOExceptions import SegmentationNotSupported, ReadPropertyException, ReadPropertyMultipleException, NoResponseFromController, ApplicationNotStarted
from ...core.utils.notes import note_and_log

#------------------------------------------------------------------------------


@note_and_log
class WhoisIAm():
    """
    Define BACnet WhoIs and IAm functions.
    """

    def whois(self, *args):
        """
        Build a WhoIs request

        :param args: string built as [ <addr>] [ <lolimit> <hilimit> ] **optional**
        :returns: discoveredDevices as a defaultdict(int)

        Example::

            whois()             # WhoIs broadcast globally.  Every device will respond with an IAm
            whois('2:5')        # WhoIs looking for the device at (Network 2, Address 5)
            whois('10 1000')    # WhoIs looking for devices in the ID range (10 - 1000) 

        """
        if not self._started:
            raise ApplicationNotStarted(
                'BACnet stack not running - use startApp()')

        if args:
            args = args[0].split()
        msg = args if args else 'any'

        self._log.debug("do_whois {!r}".format(msg))

        # build a request
        request = WhoIsRequest()
        if (len(args) == 1) or (len(args) == 3):
            request.pduDestination = Address(args[0])
            del args[0]
        else:
            request.pduDestination = GlobalBroadcast()

        if len(args) == 2:
            request.deviceInstanceRangeLowLimit = int(args[0])
            request.deviceInstanceRangeHighLimit = int(args[1])
        self._log.debug("{:>12} {}".format("- request:", request))

        iocb = IOCB(request)                            # make an IOCB

        # pass to the BACnet stack
        self.this_application.request_io(iocb)

        iocb.wait()             # Wait for BACnet response

        if iocb.ioResponse:     # successful response
            apdu = iocb.ioResponse

        if iocb.ioError:        # unsuccessful: error/reject/abort
            pass

        self.discoveredDevices = self.this_application.i_am_counter
        return self.discoveredDevices

    def iam(self):
        """
        Build an IAm response.  IAm are sent in response to a WhoIs request that;  
        matches our device ID, whose device range includes us, or is a broadcast.
        Content is defined by the script (deviceId, vendor, etc...)

        :returns: bool

        Example::

            iam()
        """

        self._log.debug("do_iam")

        try:
            # build a response
            request = IAmRequest()
            request.pduDestination = GlobalBroadcast()

            # fill the response with details about us (from our device object)
            request.iAmDeviceIdentifier = self.this_device.objectIdentifier
            request.maxAPDULengthAccepted = self.this_device.maxApduLengthAccepted
            request.segmentationSupported = self.this_device.segmentationSupported
            request.vendorID = self.this_device.vendorIdentifier
            self._log.debug("{:>12} {}".format("- request:", request))

            iocb = self.this_application.request(
                request)       # pass to the BACnet stack
            iocb.wait()
            return True

        except Exception as error:
            self._log.error("exception: {!r}".format(error))
            return False




#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
Read.py - creation of ReadProperty and ReadPropertyMultiple requests

    Used while defining an app:
    Example::

        class BasicScript(WhoisIAm, ReadProperty)

    Class::

        ReadProperty()
            def read()
            def readMultiple()

'''

#--- standard Python modules ---

#--- 3rd party modules ---
from bacpypes.debugging import bacpypes_debugging

from bacpypes.pdu import Address
from bacpypes.object import get_object_class, get_datatype
from bacpypes.apdu import PropertyReference, ReadAccessSpecification, \
    ReadPropertyRequest, ReadPropertyMultipleRequest, RejectReason, AbortReason, RejectPDU, AbortPDU

from bacpypes.basetypes import PropertyIdentifier
from bacpypes.apdu import ReadPropertyMultipleACK, ReadPropertyACK
from bacpypes.primitivedata import Unsigned
from bacpypes.constructeddata import Array
from bacpypes.iocb import IOCB
from bacpypes.core import deferred

#--- this application's modules ---
from .IOExceptions import ReadPropertyException, ReadPropertyMultipleException, NoResponseFromController, ApplicationNotStarted, UnrecognizedService, SegmentationNotSupported, UnknownPropertyError, UnknownObjectError

from ..utils.notes import note_and_log
#------------------------------------------------------------------------------


@note_and_log
class ReadProperty():
    """
    Defines BACnet Read functions: readProperty and readPropertyMultiple.
    Data exchange is made via a Queue object
    A timeout of 10 seconds allows detection of invalid device or communciation errors.
    """

    def read(self, args, arr_index=None, vendor_id=0, bacoid=None):
        """
        Build a ReadProperty request, wait for the answer and return the value

        :param args: String with <addr> <type> <inst> <prop> [ <indx> ]
        :returns: data read from device (str representing data like 10 or True)

        *Example*::

            import BAC0
            myIPAddr = '192.168.1.10/24'
            bacnet = BAC0.connect(ip = myIPAddr)
            bacnet.read('2:5 analogInput 1 presentValue')

        Requests the controller at (Network 2, address 5) for the presentValue of
        its analog input 1 (AI:1).
        """
        if not self._started:
            raise ApplicationNotStarted(
                'BACnet stack not running - use startApp()')

        args_split = args.split()
        
        self.log_title("Read property",args_split)
                       
        vendor_id = vendor_id
        bacoid = bacoid

        try:
            # build ReadProperty request
            iocb = IOCB(self.build_rp_request(
                args_split, arr_index=arr_index, vendor_id=vendor_id, bacoid=bacoid))
            # pass to the BACnet stack
            deferred(self.this_application.request_io, iocb)
            self._log.debug("{:<20} {!r}".format('iocb', iocb))

        except ReadPropertyException as error:
            # construction error
            self._log.exception("exception: {!r}".format(error))

        iocb.wait()             # Wait for BACnet response

        if iocb.ioResponse:     # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, ReadPropertyACK):               # expecting an ACK
                self._log.warning("Not an ack, see debug for more infos.")
                self._log.debug("Not an ack. | APDU : {} / {}".format((apdu, type(apdu))))
                return

            # find the datatype
            datatype = get_datatype(
                apdu.objectIdentifier[0], apdu.propertyIdentifier, vendor_id=vendor_id)
            if not datatype:
                raise TypeError("unknown datatype")

            # special case for array parts, others are managed by cast_out
            if issubclass(datatype, Array) and (apdu.propertyArrayIndex is not None):
                if apdu.propertyArrayIndex == 0:
                    value = apdu.propertyValue.cast_out(Unsigned)
                else:
                    value = apdu.propertyValue.cast_out(datatype.subtype)
            else:
                value = apdu.propertyValue.cast_out(datatype)

            self._log.info(
                "{:<20} {:<20}".format(
                'value',
                'datatype'))
            self._log.info(
                "{!r:<20} {!r:<20}".format(
                value,
                datatype))
            return value

        if iocb.ioError:        # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            if reason == 'segmentationNotSupported':
                self._log.warning(
                    "Segmentation not supported... will read properties one by one...")
                self._log.debug("The Request was : {}".format(args_split))
                value = self._split_the_read_request(args, arr_index)
                return value
            else:
                if reason == 'unknownProperty':
                    self._log.warning('Unknown property {}'.format(args))
                    if 'description' in args:
                        return ''
                    else:
                        raise UnknownPropertyError(
                            'Unknown property {}'.format(args))
                elif reason == 'unknownObject':
                    self._log.warning('Unknown object {}'.format(args))
                    raise UnknownObjectError(
                        'Unknown object {}'.format(args))
                else:
                    # Other error... consider NoResponseFromController (65)
                    # even if the realy reason is another one
                    raise NoResponseFromController(
                        "APDU Abort Reason : {}".format(reason))

    def _split_the_read_request(self, args, arr_index):
        """
        When a device doesn't support segmentation, this function
        will split the request according to the length of the
        predicted result which can be known when readin the array_index
        number 0.

        This can be a very long process as some devices count a large
        number of properties without supporting segmentation
        (FieldServers are a good example)
        """
        objlist = []
        nmbr_obj = self.read(args, arr_index=0)
        for i in range(1, nmbr_obj+1):
            objlist.append(self.read(
                args, arr_index=i))
        return objlist

    def readMultiple(self, args):
        """ Build a ReadPropertyMultiple request, wait for the answer and return the values

        :param args: String with <addr> ( <type> <inst> ( <prop> [ <indx> ] )... )...
        :returns: data read from device (str representing data like 10 or True)

        *Example*::

            import BAC0
            myIPAddr = '192.168.1.10/24'
            bacnet = BAC0.connect(ip = myIPAddr)
            bacnet.readMultiple('2:5 analogInput 1 presentValue units')

        Requests the controller at (Network 2, address 5) for the (presentValue and units) of
        its analog input 1 (AI:1).
        """
        if not self._started:
            raise ApplicationNotStarted(
                'BACnet stack not running - use startApp()')

        args = args.split()
        values = []

        self.log_title("Read Multiple",args)

        try:
            # build an ReadPropertyMultiple request
            iocb = IOCB(self.build_rpm_request(args))
            # pass to the BACnet stack
            deferred(self.this_application.request_io, iocb)
            self._log.debug("{:<20} {!r}".format('iocb', iocb))

        except ReadPropertyMultipleException as error:
            # construction error
            self._log.exception("exception: {!r}".format(error))

        iocb.wait()             # Wait for BACnet response

        if iocb.ioResponse:     # successful response
            apdu = iocb.ioResponse

            if not isinstance(apdu, ReadPropertyMultipleACK):       # expecting an ACK
                self._log.debug("{:<20}".format("not an ack"))
                self._log.warning("Not an Ack. | APDU : {} / {}".format((apdu, type(apdu))))
                return

            # loop through the results
            for result in apdu.listOfReadAccessResults:
                # here is the object identifier
                objectIdentifier = result.objectIdentifier
                               
                self.log_subtitle('{!r} : {!r}'.format(objectIdentifier[0],objectIdentifier[1]), width=114)
                self._log.info(
                    "{:<20} {:<20} {:<30} {:<20}".format(
                    'propertyIdentifier',
                    'propertyArrayIndex',
                    'value',
                    'datatype'))
                self._log.info(
                    "-"*114)

                # now come the property values per object
                for element in result.listOfResults:
                    # get the property and array index
                    propertyIdentifier = element.propertyIdentifier
                    propertyArrayIndex = element.propertyArrayIndex
                    
                    readResult = element.readResult 

                    if readResult.propertyAccessError is not None:
                        self._log.debug('Property Access Error for {}'.format(
                                readResult.propertyAccessError))
                        values.append(None)
                    else:
                        # here is the value
                        propertyValue = readResult.propertyValue

                        # find the datatype
                        datatype = get_datatype(
                            objectIdentifier[0], propertyIdentifier)

                        if not datatype:
                            raise TypeError("unknown datatype")

                        # special case for array parts, others are managed by cast_out
                        if issubclass(datatype, Array) and (propertyArrayIndex is not None):
                            if propertyArrayIndex == 0:
                                value = propertyValue.cast_out(Unsigned)
                            else:
                                value = propertyValue.cast_out(
                                    datatype.subtype)
                        else:
                            value = propertyValue.cast_out(datatype)
                            

                        self._log.info(
                            "{!r:<20} {!r:<20} {!r:<30} {!r:<20}".format(
                            propertyIdentifier,
                            propertyArrayIndex,
                            value,
                            datatype))                        
                        values.append(value)

            return values

        if iocb.ioError:        # unsuccessful: error/reject/abort
            apdu = iocb.ioError
            reason = find_reason(apdu)
            self._log.warning("APDU Abort Reject Reason : {}".format(reason))
            self._log.debug("The Request was : {}".format(args))
            if reason == 'unrecognizedService':
                raise UnrecognizedService()
            elif reason == 'segmentationNotSupported':
                raise SegmentationNotSupported()
            elif reason == 'unknownObject':
                self._log.warning('Unknown object {}'.format(args))
                raise UnknownObjectError('Unknown object {}'.format(args))
            elif reason == 'unknownProperty':
                self._log.warning('Unknown property {}'.format(args))
                values.append("")
                return values
            else:
                self._log.warning("No response from controller")
                values.append("")
                return values

    def build_rp_request(self, args, arr_index=None, vendor_id=0, bacoid=None):
        addr, obj_type, obj_inst, prop_id = args[:4]
        vendor_id = vendor_id
        bacoid = bacoid

        if obj_type.isdigit():
            obj_type = int(obj_type)
        elif not get_object_class(obj_type):
            raise ValueError("unknown object type")

        obj_inst = int(obj_inst)

        if prop_id.isdigit():
            prop_id = int(prop_id)
        datatype = get_datatype(obj_type, prop_id, vendor_id=vendor_id)
        if not datatype:
            raise ValueError("invalid property for object type")

        # build a request
        request = ReadPropertyRequest(
            objectIdentifier=(obj_type, obj_inst),
            propertyIdentifier=prop_id,
            propertyArrayIndex=arr_index,
        )
        request.pduDestination = Address(addr)

        if len(args) == 5:
            request.propertyArrayIndex = int(args[4])
        self._log.debug("{:<20} {!r}".format(
            'REQUEST', request))
        return request

    def build_rpm_request(self, args):
        """
        Build request from args
        """
        i = 0
        addr = args[i]
        i += 1

        read_access_spec_list = []
        while i < len(args):
            obj_type = args[i]
            i += 1

            if obj_type.isdigit():
                obj_type = int(obj_type)
            elif not get_object_class(obj_type):
                raise ValueError("unknown object type")

            obj_inst = int(args[i])
            i += 1

            prop_reference_list = []
            while i < len(args):
                prop_id = args[i]
                if prop_id not in PropertyIdentifier.enumerations:
                    break

                i += 1
                if prop_id in ('all', 'required', 'optional'):
                    pass
                else:
                    datatype = get_datatype(obj_type, prop_id)
                    if not datatype:
                        raise ValueError(
                            "invalid property for object type : {} | {}".format(
                                (obj_type, prop_id)))

                # build a property reference
                prop_reference = PropertyReference(propertyIdentifier=prop_id)

                # check for an array index
                if (i < len(args)) and args[i].isdigit():
                    prop_reference.propertyArrayIndex = int(args[i])
                    i += 1

                prop_reference_list.append(prop_reference)

            if not prop_reference_list:
                raise ValueError("provide at least one property")

            # build a read access specification
            read_access_spec = ReadAccessSpecification(
                objectIdentifier=(obj_type, obj_inst),
                listOfPropertyReferences=prop_reference_list)

            read_access_spec_list.append(read_access_spec)

        if not read_access_spec_list:
            raise RuntimeError(
                "at least one read access specification required")

        # build the request
        request = ReadPropertyMultipleRequest(
            listOfReadAccessSpecs=read_access_spec_list)
        request.pduDestination = Address(addr)
        self._log.debug("{:<20} {!r}".format(
            'REQUEST', request))
        return request


def find_reason(apdu):
    if apdu.pduType == RejectPDU.pduType:
        reasons = RejectReason.enumerations
    elif apdu.pduType == AbortPDU.pduType:
        reasons = AbortReason.enumerations
    else:
        if apdu.errorCode and apdu.errorClass:
            return '{}'.format(apdu.errorCode)
        else:
            raise ValueError('Cannot find reason...')
    code = apdu.apduAbortRejectReason
    try:
        return [k for k, v in reasons.items() if v == code][0]
    except IndexError:
        return code



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
Complete Script - extended version of Lite Script

As everything is handled by the BasicScript, select the additional features you want::

    # Create a class that implements a basic script with read and write functions
    from BAC0.scripts.BasicScript import BasicScript
    from BAC0.core.io.Read import ReadProperty
    from BAC0.core.io.Write import WriteProperty
    class ReadWriteScript(BasicScript,ReadProperty,WriteProperty)

Once the class is created, create the local object and use it::

    bacnet = ReadWriteScript(localIPAddr = '192.168.1.10')
    bacnet.read('2:5 analogInput 1 presentValue)

'''
#--- standard Python modules ---
from datetime import datetime

import pandas as pd


#--- 3rd party modules ---
from bokeh.application import Application

#--- this application's modules ---
from ..scripts.Lite import Lite

from ..core.io.IOExceptions import BokehServerCantStart, NoResponseFromController, UnrecognizedService
from ..core.utils.notes import note_and_log

from ..web.BokehRenderer import DevicesTableHandler, DynamicPlotHandler, NotesTableHandler
from ..web.BokehServer import Bokeh_Worker
from ..web.FlaskServer import FlaskServer

#------------------------------------------------------------------------------


class Stats_Mixin():
    """
    Statistics used by Flask App
    """
    @property
    def number_of_devices(self):
        s = []
        [s.append(x) for x in self.whois_answer[0].items() if x[1] > 0]
        return len(s)

    @property
    def number_of_registered_trends(self):
        if self.trends:
            return len(self.trends)
        else:
            return 0

    def number_of_devices_per_network(self):
        total = float(self.number_of_devices)
        if total == 0:
            return (['No Devices'], ['0'], ['0%%'])
        labels = ['IP']
        series_pct = ['%.2f %%' %
                      (len(self.network_stats['ip_devices'])/total * 100)]
        series = [len(self.network_stats['ip_devices'])/total * 100]
        for each in (self.network_stats['mstp_map'].keys()):
            labels.append('MSTP #%s' % each)
            series_pct.append('%.2f %%' % (
                len(self.network_stats['mstp_map'][each])/total * 100))
            series.append(
                len(self.network_stats['mstp_map'][each])/total * 100)
        return (labels, series, series_pct)

    @property
    def network_stats(self):
        """
        Used by Flask to show informations on the network
        """
        statistics = {}
        mstp_networks = []
        mstp_map = {}
        ip_devices = []
        bacoids = []
        mstp_devices = []
        for address, bacoid in self.whois_answer[0].keys():
            if ':' in address:
                net, mac = address.split(':')
                mstp_networks.append(net)
                mstp_devices.append(mac)
                try:
                    mstp_map[net].append(mac)
                except KeyError:
                    mstp_map[net] = []
                    mstp_map[net].append(mac)
            else:
                net = 'ip'
                mac = address
                ip_devices.append(address)
            bacoids.append((bacoid, address))
        mstpnetworks = sorted(set(mstp_networks))
        statistics['mstp_networks'] = mstpnetworks
        statistics['ip_devices'] = sorted(ip_devices)
        statistics['bacoids'] = sorted(bacoids)
        statistics['mstp_map'] = mstp_map
        statistics['timestamp'] = str(datetime.now())
        statistics['number_of_devices'] = self.number_of_devices
        statistics['number_of_registered_devices'] = len(
            self.registered_devices)
        statistics['print_mstpnetworks'] = print_list(mstpnetworks)
        return statistics


@note_and_log
class Complete(Lite, Stats_Mixin):
    """
    Build a BACnet application to accept read and write requests.
    [Basic Whois/IAm functions are implemented in parent BasicScript class.]
    Once created, execute a whois() to build a list of available controllers.
    Initialization requires information on the local device.

    :param ip='127.0.0.1': Address must be in the same subnet as the BACnet network
        [BBMD and Foreign Device - not supported]

    :param bokeh_server: (boolean) If set to false, will prevent Bokeh server
        from being started. Can help troubleshoot issues with Bokeh. By default,
        set to True.
    """

    def __init__(self, ip=None,
                 bbmdAddress=None, bbmdTTL=0,
                 bokeh_server=True, flask_port=8111):
        Lite.__init__(self, ip=ip, bbmdAddress=bbmdAddress, bbmdTTL=bbmdTTL)
        self.flask_port = flask_port
        if bokeh_server:
            self.start_bokeh()
            self.FlaskServer.start()
        else:
            self._log.warning(
                'Bokeh server not started. Trend feature will not work')

    @property
    def devices(self):
        lst = []
        for device in list(self.discoveredDevices):
            try:
                deviceName, vendorName = self.readMultiple(
                    '{} device {} objectName vendorName'.format(device[0], device[1]))
            except UnrecognizedService:
                deviceName = self.read(
                    '{} device {} objectName'.format(device[0], device[1]))
                vendorName = self.read(
                    '{} device {} vendorName'.format(device[0], device[1]))
            except NoResponseFromController:
                self._log.info('No response from {}'.format(device))
                continue
            lst.append((deviceName, vendorName, device[0], device[1]))
        df = pd.DataFrame(lst, columns=[
            'Name', 'Manufacturer', 'Address', ' Device ID']).set_index('Name')
        try:
            return df.sort_values('Address')
        except AttributeError:
            return df

    def start_bokeh(self):
        try:
            self.note('Starting Bokeh Serve')
            # Need to create the device document here
            devHandler = DevicesTableHandler(self)
            dev_app = Application(devHandler)
            trendHandler = DynamicPlotHandler(self)
            notesHandler = NotesTableHandler(self)
            self.trend_app = Application(trendHandler)
            self.notes_app = Application(notesHandler)
            self.bk_worker = Bokeh_Worker(
                dev_app, self.trend_app, self.notes_app, self.localIPAddr)
            self.FlaskServer = FlaskServer(
                network=self, port=self.flask_port, ip=self.localIPAddr)
            self.bk_worker.start()
            self.bokehserver = True
            print('Server started : http://localhost:{}'.format(self.flask_port))

        except OSError as error:
            self.bokehserver = False
            self._log.error(
                '[bokeh serve] required for trending (controller.chart) features')
            self._log.error(error)

        except RuntimeError as rterror:
            self.bokehserver = False
            self._log.warning('Server already running')

        except BokehServerCantStart:
            self.bokehserver = False
            self._log.error('No Bokeh Server - controller.chart not available')

    def __repr__(self):
        return 'Bacnet Network using ip {} with device id {} | Featuring Bokeh and Pandas'.format(self.localIPAddr, self.Boid)



#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
sql.py - 
'''

#--- standard Python modules ---
import pickle
import os.path

#--- 3rd party modules ---
import sqlite3

try:
    import pandas as pd
    from pandas.io import sql
    try:
        from pandas import Timestamp
    except ImportError:
        from pandas.lib import Timestamp
    _PANDAS = True
except ImportError:
    _PANDAS = False
#--- this application's modules ---

#------------------------------------------------------------------------------

class SQLMixin(object):
    """
    Use SQL to persist a device's contents.  By saving the device contents to an SQL 
    database, you can work with the device's data while offline, or while the device 
    is not available. 
    """

    def dev_properties_df(self):
        dic = self.properties.asdict.copy()
        #dic.pop('charts', None)
        dic.pop('network', None)
        dic.pop('pss', None)
        #dic.pop('serving_chart', None)
        return dic
        
    
    def points_properties_df(self):
        """
        Return a dictionary of point/point_properties in preparation for storage in SQL.
        """
        pprops = {}
        for each in self.points:
            p = each.properties.asdict.copy()
            #p.pop('charts', None)
            p.pop('device', None)
            p.pop('network', None)
            p.pop('simulated', None)
            p.pop('overridden', None)
            pprops[each.properties.name] = p

        df = pd.DataFrame(pprops)
        return df


    def backup_histories_df(self):
        """
        Build a dataframe of the point histories
        """
        backup = {}
        for point in self.points:
            if point.history.dtypes == object:
                backup[point.properties.name] = point.history.replace(['inactive', 'active'], [0, 1]).resample('1s').mean()
            else:
                backup[point.properties.name] = point.history.resample('1s').mean()

        # in some circumstances, correct : pd.DataFrame(dict([ (k,pd.Series(v)) for k,v in backup.items() ]))
        backup = pd.DataFrame(dict([ (k,pd.Series(v)) for k,v in backup.items() ]))
        return pd.DataFrame(backup)


    def save(self, filename = None):
        """
        Save the point histories to sqlite3 database.  
        Save the device object properties to a pickle file so the device can be reloaded.
        """
        if filename:
            self.properties.db_name = filename
        else:
            self.properties.db_name = self.properties.name
            
        # Does file exist? If so, append data
        if os.path.isfile('%s.db' % (self.properties.db_name)):
            #print('File exists, appending data...')

            db = sqlite3.connect('%s.db' % (self.properties.db_name))
            his = sql.read_sql('select * from "%s"' % 'history', db)  
            his.index = his['index'].apply(Timestamp)
            last = his.index[-1]
            df_to_backup = self.backup_histories_df()[last:]
            db.close()
            
        else:
            self._log.debug('Creating a new backup database')
            df_to_backup = self.backup_histories_df()
        
        cnx = sqlite3.connect('%s.db' % (self.properties.db_name))
    
        # DataFrames that will be saved to SQL
        sql.to_sql(df_to_backup, name='history', con=cnx, index_label = 'index', index = True, if_exists = 'append')

        prop_backup = {}
        prop_backup['device'] = self.dev_properties_df()
        prop_backup['points'] = self.points_properties_df()
        with open( "%s.bin"  % self.properties.db_name, "wb" ) as file:
            pickle.dump(prop_backup, file)
                
        #print('%s saved to disk' % self.properties.db_name)
        

    def points_from_sql(self, db):
        """
        Retrieve point list from SQL database
        """
        points = sql.read_sql("SELECT * FROM history;", db) 
        return list(points.columns.values)[1:]
        

    def his_from_sql(self, db, point):
        """
        Retrive point histories from SQL database
        """
        his = sql.read_sql('select * from "%s"' % 'history', db)  
        his.index = his['index'].apply(Timestamp)
        return his.set_index('index')[point]
        

    def value_from_sql(self, db, point):
        """
        Take last known value as the value
        """
        return self.his_from_sql(db, point).last_valid_index()          
        

    def read_point_prop(self, device_name, point):
        """
        Points properties retrieved from pickle
        """
        with open( "%s.bin" % device_name, "rb" ) as file:        
            return pickle.load(file)['points'][point]


    def read_dev_prop(self, device_name):
        """
        Device properties retrieved from pickle
        """
        with open( "%s.bin" % device_name, "rb" ) as file:
            return pickle.load(file)['device']            
    


#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
Points.py - Definition of points so operations on Read results are more convenient.
'''

#--- standard Python modules ---
from datetime import datetime
from collections import namedtuple
import time

#--- 3rd party modules ---
try:
    import pandas as pd
    from pandas.io import sql
    try:
        from pandas import Timestamp
    except ImportError:
        from pandas.lib import Timestamp
    _PANDAS = True
except ImportError:
    _PANDAS = False

#--- this application's modules ---
from ...tasks.Poll import SimplePoll as Poll
from ...tasks.Match import Match, Match_Value
from ..io.IOExceptions import NoResponseFromController

#------------------------------------------------------------------------------


class PointProperties(object):
    """
    A container for point properties.
    """

    def __init__(self):
        self.device = None
        self.name = None
        self.type = None
        self.address = None
        self.description = None
        self.units_state = None
        self.simulated = (False, None)
        self.overridden = (False, None)

    def __repr__(self):
        return '%s' % self.asdict

    @property
    def asdict(self):
        return self.__dict__

#------------------------------------------------------------------------------


class Point():
    """
    Represents a device BACnet point.  Used to NumericPoint, BooleanPoint and EnumPoints.

    Each point implements a history feature. Each time the point is read, its value (with timestamp)
    is added to a history table. Histories capture the changes to point values over time.
    """

    def __init__(self, device=None,
                 pointType=None,    pointAddress=None,  pointName=None,
                 description=None,  presentValue=None,  units_state=None):

        self._history = namedtuple('_history', ['timestamp', 'value'])
        self.properties = PointProperties()

        self._polling_task = namedtuple('_polling_task', ['task', 'running'])
        self._polling_task.task = None
        self._polling_task.running = False

        self._match_task = namedtuple('_match_task', ['task', 'running'])
        self._match_task.task = None
        self._match_task.running = False

        self._history.timestamp = []
        self._history.value = []
        self._history.value.append(presentValue)
        self._history.timestamp.append(datetime.now())

        self.properties.device = device
        self.properties.name = pointName
        self.properties.type = pointType
        self.properties.address = pointAddress

        self.properties.description = description
        self.properties.units_state = units_state
        self.properties.simulated = (False, 0)
        self.properties.overridden = (False, 0)

    @property
    def value(self):
        """
        Retrieve value of the point
        """
        try:
            res = self.properties.device.properties.network.read('{} {} {} presentValue'.format(
                self.properties.device.properties.address, self.properties.type, str(self.properties.address)))
            self._trend(res)
        except Exception:
            raise Exception(
                'Problem reading : {}'.format(self.properties.name))

        return res

    def _trend(self, res):
        self._history.timestamp.append(datetime.now())
        self._history.value.append(res)

    @property
    def units(self):
        """
        Should return units
        """
        raise Exception('Must be overridden')

    @property
    def lastValue(self):
        """
        returns: last value read
        """
        if _PANDAS:
            return self.history.dropna().iloc[-1]
        else:
            return self._history.value[-1]

    @property
    def history(self):
        """
        returns : (pd.Series) containing timestamp and value of all readings
        """
        if not _PANDAS:
            return dict(zip(self._history.timestamp, self._history.value))
        his_table = pd.Series(self._history.value,
                              index=self._history.timestamp)
        his_table.name = (
            '{}/{}').format(self.properties.device.properties.name, self.properties.name)
        his_table.units = self.properties.units_state
        if self.properties.name in self.properties.device.binary_states:
            his_table.states = 'binary'
        elif self.properties.name in self.properties.device.multi_states:
            his_table.states = 'multistates'
        else:
            his_table.states = 'analog'
        his_table.description = self.properties.description

        his_table.datatype = self.properties.type
        return his_table

    def clear_history(self):
        self._history.timestamp = []
        self._history.value = []

    def chart(self, remove=False):
        """
        Add point to the bacnet trending list
        """
        if remove:
            self.properties.device.properties.network.remove_trend(self)
        else:
            self.properties.device.properties.network.add_trend(self)

    def __getitem__(self, key):
        """
        Way to get points... presentValue, status, flags, etc...

        :param key: state
        :returns: list of enum states
        """
        if str(key).lower() in ['unit', 'units', 'state', 'states']:
            key = 'units_state'
        try:
            return getattr(self.properties, key)
        except AttributeError:
            raise ValueError('Wrong property')

    def write(self, value, *, prop='presentValue', priority=''):
        """
        Write to present value of a point

        :param value: (float) numeric value
        :param prop: (str) property to write. Default = presentValue
        :param priority: (int) priority to which write.

        """
        if priority != '':
            if isinstance(float(priority), float)\
                    and float(priority) >= 1\
                    and float(priority) <= 16:
                priority = '- {}'.format(priority)
            else:
                raise ValueError('Priority must be a number between 1 and 16')

        try:
            self.properties.device.properties.network.write(
                '{} {} {} {} {} {}'.format(
                    self.properties.device.properties.address,
                    self.properties.type,
                    self.properties.address,
                    prop,
                    value,
                    priority))
        except Exception:
            raise NoResponseFromController()

        # Read after the write so history gets updated.
        self.value

    def default(self, value):
        self.write(value, prop='relinquishDefault')

    def sim(self, value, *, force=False):
        """
        Simulate a value.  Sets the Out_Of_Service property- to disconnect the point from the
        controller's control.  Then writes to the Present_Value.
        The point name is added to the list of simulated points (self.simPoints)

        :param value: (float) value to simulate
        """
        if self.properties.simulated[0] \
                and self.properties.simulated[1] == value \
                and force == False:
            pass
        else:
            self.properties.device.properties.network.sim('{} {} {} presentValue {}'.format(
                self.properties.device.properties.address, self.properties.type, str(self.properties.address), str(value)))
            self.properties.simulated = (True, value)

    def out_of_service(self):
        """
        Sets the Out_Of_Service property [to True].
        """
        self.properties.device.properties.network.out_of_service('{} {} {}'.format(
            self.properties.device.properties.address, self.properties.type, str(self.properties.address)))
        self.properties.simulated = (True, None)

    def release(self):
        """
        Clears the Out_Of_Service property [to False] - so the controller regains control of the point.
        """
        self.properties.device.properties.network.release('{} {} {}'.format(
            self.properties.device.properties.address, self.properties.type, str(self.properties.address)))
        self.properties.simulated = (False, None)

    def ovr(self, value):
        self.write(value, priority=8)
        self.properties.overridden = (True, value)

    def auto(self):
        self.write('null', priority=8)
        self.properties.overridden = (False, 0)

    def _setitem(self, value):
        """
        Called by _set, will trigger right function depending on
        point type to write to the value and make tests.
        This is default behaviour of the point  :
        AnalogValue are written to
        AnalogOutput are overridden
        """
        if 'Value' in self.properties.type:
            if str(value).lower() == 'auto':
                raise ValueError(
                    'Value was not simulated or overridden, cannot release to auto')
            # analog value must be written to
            self.write(value)

        elif 'Output' in self.properties.type:
            # analog output must be overridden
            if str(value).lower() == 'auto':
                self.auto()
            else:
                self.ovr(value)
        else:
            # input are left... must be simulated
            if str(value).lower() == 'auto':
                self.release()
            else:
                self.sim(value)

    def _set(self, value):
        """
        Allows the syntax:
            device['point'] = value
        """
        raise Exception('Must be overridden')

    def poll(self, command='start', *, delay=10):
        """
        Poll a point every x seconds (delay=x sec)
        Stopped by using point.poll('stop') or .poll(0) or .poll(False)
        or by setting a delay = 0
        """
        if str(command).lower() == 'stop' \
                or command == False \
                or command == 0 \
                or delay == 0:

            if isinstance(self._polling_task.task, Poll):
                self._polling_task.task.stop()
                self._polling_task.task = None
                self._polling_task.running = False

        elif self._polling_task.task is None:
            self._polling_task.task = Poll(self, delay=delay)
            self._polling_task.task.start()
            self._polling_task.running = True

        elif self._polling_task.running:
            self._polling_task.task.stop()
            self._polling_task.running = False
            self._polling_task.task = Poll(self, delay=delay)
            self._polling_task.task.start()
            self._polling_task.running = True

        else:
            raise RuntimeError('Stop polling before redefining it')

    def match(self, point, *, delay=5):
        """
        This allow functions like :
            device['status'].match('command')

        A fan status for example will follow the command...
        """
        if self._match_task.task is None:
            self._match_task.task = Match(
                command=point, status=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay > 0:
            self._match_task.task.stop()
            self._match_task.running = False
            time.sleep(1)

            self._match_task.task = Match(
                command=point, status=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay == 0:
            self._match_task.task.stop()
            self._match_task.running = False

        else:
            raise RuntimeError('Stop task before redefining it')

    def match_value(self, value, *, delay=5):
        """
        This allow functions like :
            device['point'].match('value')

        A sensor will follow a calculation...
        """
        if self._match_task.task is None:
            self._match_task.task = Match_Value(
                value=value, point=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay > 0:
            self._match_task.task.stop()
            self._match_task.running = False
            time.sleep(1)

            self._match_task.task = Match_Value(
                value=value, point=self, delay=delay)
            self._match_task.task.start()
            self._match_task.running = True

        elif self._match_task.running and delay == 0:
            self._match_task.task.stop()
            self._match_task.running = False

        else:
            raise RuntimeError('Stop task before redefining it')

    def __len__(self):
        """
        Length of a point = # of history records
        """
        return len(self.history)


#------------------------------------------------------------------------------

class NumericPoint(Point):
    """
    Representation of a Numeric value
    """

    def __init__(self, device=None,
                 pointType=None,    pointAddress=None,  pointName=None,
                 description=None,  presentValue=None,  units_state=None):

        Point.__init__(self, device=device,
                       pointType=pointType,     pointAddress=pointAddress,  pointName=pointName,
                       description=description, presentValue=presentValue,  units_state=units_state)

    @property
    def units(self):
        return self.properties.units_state

    def _set(self, value):
        if str(value).lower() == 'auto':
            self._setitem(value)
        else:
            try:
                if isinstance(value, Point):
                    value = value.lastValue
                val = float(value)
                if isinstance(val, float):
                    self._setitem(value)
            except:
                raise ValueError('Value must be numeric')

    def __repr__(self):
        return '{}/{} : {:.2f} {}'.format(self.properties.device.properties.name, self.properties.name, self.lastValue, self.properties.units_state)

    def __add__(self, other):
        return self.value + other

    def __sub__(self, other):
        return self.value - other

    def __mul__(self, other):
        return self.value * other

    def __truediv__(self, other):
        return self.value / other

    def __lt__(self, other):
        return self.value < other

    def __le__(self, other):
        return self.value <= other

    def __eq__(self, other):
        return self.value == other

    def __gt__(self, other):
        return self.value > other

    def __ge__(self, other):
        return self.value >= other

#------------------------------------------------------------------------------


class BooleanPoint(Point):
    """
    Representation of a Boolean value
    """

    def __init__(self, device=None,
                 pointType=None,    pointAddress=None,  pointName=None,
                 description=None,  presentValue=None,  units_state=None):

        Point.__init__(self, device=device,
                       pointType=pointType,     pointAddress=pointAddress,  pointName=pointName,
                       description=description, presentValue=presentValue,  units_state=units_state)

    @property
    def value(self):
        """
        Read the value from BACnet network
        """
        try:
            res = self.properties.device.properties.network.read('{} {} {} presentValue'.format(
                self.properties.device.properties.address, self.properties.type, str(self.properties.address)))
            self._trend(res)

        except Exception:
            raise Exception(
                'Problem reading : {}'.format(self.properties.name))

        if res == 'inactive':
            self._key = 0
            self._boolKey = False
        else:
            self._key = 1
            self._boolKey = True
        return res

    @property
    def boolValue(self):
        """
        returns : (boolean) Value
        """
        if self.lastValue == 1 or self.lastValue == 'active':
            self._key = 1
            self._boolKey = True
        else:
            self._key = 0
            self._boolKey = False
        return self._boolKey

    @property
    def units(self):
        """
        Boolean points don't have units
        """
        return None

    def _set(self, value):
        if value == True:
            self._setitem('active')
        elif value == False:
            self._setitem('inactive')
        elif str(value) in ['inactive', 'active'] or str(value).lower() == 'auto':
            self._setitem(value)
        else:
            raise ValueError(
                'Value must be boolean True, False or "active"/"inactive"')

    def __repr__(self):
        return '{}/{} : {}'.format(self.properties.device.properties.name, self.properties.name, self.boolValue)

    def __or__(self, other):
        return self.boolValue | other

    def __and__(self, other):
        return self.boolValue & other

    def __xor__(self, other):
        return self.boolValue ^ other

    def __eq__(self, other):
        return self.boolValue == other

#------------------------------------------------------------------------------


class EnumPoint(Point):
    """
    Representation of an Enumerated (multiState) value
    """

    def __init__(self, device=None,
                 pointType=None,    pointAddress=None,  pointName=None,
                 description=None,  presentValue=None,  units_state=None):

        Point.__init__(self, device=device,
                       pointType=pointType,     pointAddress=pointAddress,  pointName=pointName,
                       description=description, presentValue=presentValue,  units_state=units_state)

    @property
    def enumValue(self):
        """
        returns: (str) Enum state value
        """
        try:
            return self.properties.units_state[int(self.lastValue) - 1]
        except IndexError:
            value = 'unknown'
        except ValueError:
            value = 'NaN'
        return value

    @property
    def units(self):
        """
        Enums have 'state text' instead of units.
        """
        return None

    def _set(self, value):
        if isinstance(value, int):
            self._setitem(value)
        elif str(value) in self.properties.units_state:
            self._setitem(self.properties.units_state.index(value) + 1)
        elif str(value).lower() == 'auto':
            self._setitem('auto')
        else:
            raise ValueError(
                'Value must be integer or correct enum state : {}'.format(self.properties.units_state))

    def __repr__(self):
        # return '%s : %s' % (self.name, )
        return '{}/{} : {}'.format(self.properties.device.properties.name, self.properties.name, self.enumValue)

    def __eq__(self, other):
        return self.value == self.properties.units_state.index(other) + 1


#------------------------------------------------------------------------------

class OfflinePoint(Point):
    """
    When offline (DB state), points needs to behave in a particular way
    (we can't read on bacnet...)
    """

    def __init__(self, device, name):
        self.properties = PointProperties()
        self.properties.device = device
        dev_name = self.properties.device.properties.db_name
        props = self.properties.device.read_point_prop(dev_name, name)

        self.properties.name = props['name']
        self.properties.type = props['type']
        self.properties.address = props['address']

        self.properties.description = props['description']
        self.properties.units_state = props['units_state']
        self.properties.simulated = 'Offline'
        self.properties.overridden = 'Offline'

        if 'analog' in self.properties.type:
            self.new_state(NumericPointOffline)
        elif 'multi' in self.properties.type:
            self.new_state(EnumPointOffline)
        elif 'binary' in self.properties.type:
            self.new_state(BooleanPointOffline)
        else:
            raise TypeError('Unknown point type')

    def new_state(self, newstate):
        self.__class__ = newstate


class NumericPointOffline(NumericPoint):
    @property
    def history(self):
        his = sql.read_sql('select * from "{}"'.format(
            'history', self.properties.device.db))
        his.index = his['index'].apply(Timestamp)
        return his.set_index('index')[self.properties.name]

    @property
    def value(self):
        """
        Take last known value as the value
        """
        try:
            value = self.lastValue
        except IndexError:
            value = 65535
        return value

    def write(self, value, *, prop='presentValue', priority=''):
        raise OfflineException('Must be online to write')

    def sim(self, value, *, prop='presentValue', priority=''):
        raise OfflineException('Must be online to write')

    def release(self, value, *, prop='presentValue', priority=''):
        raise OfflineException('Must be online to write')

    @property
    def units(self):
        return self.properties.units_state

    def _set(self, value):
        raise OfflineException('Must be online to write')

    def __repr__(self):
        return '{}/{} : {:.2f} {}'.format(self.properties.device.properties.name, self.properties.name, self.value, self.properties.units_state)


class BooleanPointOffline(BooleanPoint):
    @property
    def history(self):
        his = sql.read_sql('select * from "{}"'.format(
            'history', self.properties.device.db))
        his.index = his['index'].apply(Timestamp)
        return his.set_index('index')[self.properties.name]

    @property
    def value(self):
        try:
            value = self.lastValue
        except IndexError:
            value = 'NaN'
        return value

    def _set(self, value):
        raise OfflineException('Point must be online to write')

    def write(self, value, *, prop='presentValue', priority=''):
        raise OfflineException('Must be online to write')

    def sim(self, value, *, prop='presentValue', priority=''):
        raise OfflineException('Must be online to write')

    def release(self, value, *, prop='presentValue', priority=''):
        raise OfflineException('Must be online to write')


class EnumPointOffline(EnumPoint):
    @property
    def history(self):
        his = sql.read_sql('select * from "{}"'.format(
            'history', self.properties.device.db))
        his.index = his['index'].apply(Timestamp)
        return his.set_index('index')[self.properties.name]

    @property
    def value(self):
        """
        Take last known value as the value
        """
        try:
            value = self.lastValue
        except IndexError:
            value = 'NaN'
        except ValueError:
            value = 'NaN'
        return value

    @property
    def enumValue(self):
        """
        returns: (str) Enum state value
        """
        try:
            value = self.properties.units_state[int(self.lastValue) - 1]
        except IndexError:
            value = 'unknown'
        except ValueError:
            value = 'NaN'
        return value

    def _set(self, value):
        raise OfflineException('Point must be online to write')

    def write(self, value, *, prop='presentValue', priority=''):
        raise OfflineException('Must be online to write')

    def sim(self, value, *, prop='presentValue', priority=''):
        raise OfflineException('Must be online to write')

    def release(self, value, *, prop='presentValue', priority=''):
        raise OfflineException('Must be online to write')


class OfflineException(Exception):
    pass



"""
Setup.py
"""
from setuptools import setup
from BAC0 import infos

requirements = [
    'bacpypes',
]

setup(name='BAC0',
      version=infos.__version__,
      description='BACnet Scripting Framework for testing DDC Controls',
      author=infos.__author__,
      author_email=infos.__email__,
      url=infos.__url__,
      download_url=infos.__download_url__,
      keywords=['bacnet', 'building', 'automation', 'test'],
      packages=[
          'BAC0',
          'BAC0.core',
          'BAC0.core.app',
          'BAC0.core.io',
          'BAC0.core.functions',
          'BAC0.core.devices',
          'BAC0.core.devices.mixins',
          'BAC0.core.utils',
          'BAC0.core.proprietary_objects',
          'BAC0.scripts',
          'BAC0.tasks',
          'BAC0.web',
          'BAC0.sql'
      ],
      include_package_data=True,       
      requires=requirements,
      install_requires=requirements,
      test_suite="tests",
      long_description=open('README.rst').read(),
      classifiers=[
          "Development Status :: 4 - Beta",
          "Intended Audience :: Developers",
          "Operating System :: OS Independent",
          "Programming Language :: Python",
          "Topic :: Software Development :: Libraries :: Python Modules",
          "Topic :: System :: Networking",
          "Topic :: Utilities",
      ],)



#!/usr/bin/env python

"""
Johnson Controls Proprietary Objects for FX/FEC Line
"""
from bacpypes.primitivedata import Boolean
from bacpypes.object import BinaryValueObject, Property, register_object_type

#
#   Vendor TECSupOnline Value Object Type
#

class TECSupOnline(BinaryValueObject):
    """
    This class will provide to a BAC0 device referring to a Johnson Controls TEC3000
    thermostats the Supervisor Online variable. This will allow testing of remote
    occupancy schedules.
    """
    objectType = 'device'
    vendor_id = 5

    properties = [
        Property(3653, Boolean, mutable=True),
        ]

    def __init__(self, **kwargs):
        BinaryValueObject.__init__(self, **kwargs)

def register(cls, vendor_id=5):
    register_object_type(cls, vendor_id=vendor_id)




#!/usr/bin/python
# -*- coding: utf-8 -*-
#
# Copyright (C) 2015 by Christian Tremblay, P.Eng <christian.tremblay@servisys.com>
# Licensed under LGPLv3, see file LICENSE in this source tree.
#
'''
Simulate.py - simulate the value of controller I/O values
'''

#--- standard Python modules ---
#--- 3rd party modules ---
#--- this application's modules ---
from .IOExceptions import OutOfServiceNotSet, OutOfServiceSet, NoResponseFromController, ApplicationNotStarted


#------------------------------------------------------------------------------

class Simulation():
    """
    Global informations regarding simulation
    """

    def sim(self, args):
        """
        Simulate I/O points by setting the Out_Of_Service property, then doing a 
        WriteProperty to the point's Present_Value.

        :param args: String with <addr> <type> <inst> <prop> <value> [ <indx> ] [ <priority> ]

        """
        if not self._started:
            raise ApplicationNotStarted('BACnet stack not running - use startApp()')

        #with self.this_application._lock: if use lock...won't be able to call read...
        args = args.split()
        addr, obj_type, obj_inst, prop_id, value = args[:5]

        if self.read('{} {} {} outOfService'.format(addr, obj_type, obj_inst)):
            self.write('{} {} {} {} {}'.format(addr, obj_type, obj_inst, prop_id, value))
        else:
            try:
                self.write('{} {} {} outOfService True'.format(addr, obj_type, obj_inst))
            except NoResponseFromController:
                pass

            try:
                if self.read('{} {} {} outOfService'.format(addr, obj_type, obj_inst)):
                    self.write('{} {} {} {} {}'.format(addr, obj_type, obj_inst, prop_id, value))
                else:
                    raise OutOfServiceNotSet()
            except NoResponseFromController:
                pass


    def out_of_service(self, args):
        """
        Set the Out_Of_Service property so the Present_Value of an I/O may be written.

        :param args: String with <addr> <type> <inst> <prop> <value> [ <indx> ] [ <priority> ]

        """
        if not self._started:
            raise ApplicationNotStarted('BACnet stack not running - use startApp()')

        #with self.this_application._lock: if use lock...won't be able to call read...
        args = args.split()
        addr, obj_type, obj_inst = args[:3]
        try:
            self.write('{} {} {} outOfService True'.format(addr, obj_type, obj_inst))
        except NoResponseFromController:
            pass


    def release(self, args):
        """
        Set the Out_Of_Service property to False - to release the I/O point back to 
        the controller's control.

        :param args: String with <addr> <type> <inst>

        """
        if not self._started:
            raise ApplicationNotStarted('BACnet stack not running - use startApp()')

        args = args.split()
        addr, obj_type, obj_inst = args[:3]
        try:
            self.write('{} {} {} outOfService False'.format(addr, obj_type, obj_inst))
        except NoResponseFromController:
            pass

        try:
            if self.read('{} {} {} outOfService'.format(addr, obj_type, obj_inst)):
                raise OutOfServiceSet()
            else:
                pass        # Everything is ok"
        except NoResponseFromController:
            pass
